import java.awt.*;
import java.applet.*;
import java.net.*;
import java.io.*;
import java.util.Timer;
import java.util.TimerTask;

public class Log extends Applet
{
    TextField        output;
    TextArea         logWindow;
    Socket           socket;
    DataOutputStream sout;
    DataInputStream  sin;
    Timer            timer;
    // we support three granularities 1, 2, 3, for ERR, WARN, and INFO
    // XXX add constats here, and UI
    int granularity = 1;  
    // for synchronize log infos
    long timestamp = 0;  

    public void init() {
        logWindow = new TextArea( "pCache Proxy Cache Version 0.01\n", 30, 80);  
        logWindow.setEditable(false);
        add(logWindow);

       try {
          socket = new Socket( getCodeBase().getHost(), 8001); // default port
          sin = new DataInputStream( socket.getInputStream() );
          sout = new DataOutputStream( socket.getOutputStream() );
          
          Handler handler = new Handler(sin, logWindow, this);
          handler.start();
          showStatus( "connected to " + socket.getInetAddress().getHostName() + ":" + socket.getPort() );

            timer = new Timer();
            timer.scheduleAtFixedRate(new TimerTask() {
                    public void run() {
                        // send a request: a 1 byte granularity 
                        //                 and a 8 byte timestamp
                        try {
                            sout.writeByte(granularity);
                            sout.writeLong(timestamp);
                            sout.flush();
                        }
                        catch (IOException ex) {
                            logWindow.append("Applet Exception: " + ex.toString() + "\n");
                            cancel();
                        }
                    }
                }, 1000, 3000
            );
        }
        catch (IOException e) {
            showStatus( e.toString() );
        }
    }

    /* Stop the applet and close the socket */
    public void stop() {
        try {
            if (socket != null) socket.close();
        }
        catch( IOException e ) {
            logWindow.append( e.toString() + "\n" );
        }
    }
    
}

class Handler extends Thread {
    DataInputStream is;
    TextArea logWindow;
    Log log;

    public Handler(DataInputStream is, TextArea logWindow, Log log) {
        this.is = is; 
        this.logWindow = logWindow;
        this.log = log;
    }
  
    public void run() {
        try {
            String line;
            for(;;) {
                line = is.readLine();
                if (line == null) break;
                // update the timestamp
                String[] fields = line.split(" ", 3);
                if (fields.length == 3) {  // if we get it through
                    long timestamp = Long.parseLong(fields[0]);
                    // System.out.println("Updated timestamp to: " + timestamp);
                    log.timestamp = timestamp;
                }
                logWindow.append(line + "\n");
            }
        }
        catch( IOException e ) {
            logWindow.append("Applet Exception: " +  e.toString() + "\n");
        }
        finally {
            logWindow.append("Connection closed by Server\n" );
        }
    }
}
