#include <stdio.h>
#include <iostream>
#include "BitTorrTrafficIdentifier.h"
#include "Segment.h"
#include "Util.h"
#include "pCache.h"

BitTorrTrafficIdentifier::BitTorrTrafficIdentifier(char *buff, int buff_len, Conn *conn):
				TrafficIdentifier(buff, buff_len, conn), state(NO_GUESS), piecelen_final(-1)
{
	if(buff_len < 68)
		fprintf(stderr, "\n\t****\t Warning!!! less than 68 bytes passed to BitTorrTrafficIdentifier constructor.\n\n");

	this->conn = conn;
	we_are_wrong = c2s_handshaked = s2c_handshaked = false;
	BUFF_SIZE = conn->BUFSIZE;
	c2s_n_bytes_blindly_fwd = s2c_n_bytes_blindly_fwd = 0;
	do_set_rawdata_off = false;
	memcpy(hash, buff + 28, 20);

	Piece_Len_Analyzer = PieceLenAnalyzer::get_instance();

	char temp[50];
	for(int i = 0; i < 20; i++)
		sprintf( temp + 2 * i, "%02x", (unsigned char) /*note: do not use (unsigned) int (ff is printed ffffffff)*/ hash[i] );

	hash_str = temp;

	if(dbg_info_level & DBG_INF_TI_CONSTR)
		cerr << "an instance of BitTorrTrafficIdentifier was constructed for the file " << hash_str << endl;
}

int BitTorrTrafficIdentifier::c2sFilter(char **circbuff_ptr1, char *circbuff_cptr, char **circbuff_ptr3)
//note: this function must be called only when something is going from an insider to an outsider,
//		i.e. in "c2s" notion, client means the insider and server means the outsider
//*circbuff_ptr1: address of the first byte of the data in the buffer
//circbuff_cptr:  address of the begining of the buffer memory
//*circbuff_ptr3: address of the end of the data in the buffer (exclusive)
//return value:
//		positivie integer (x): forward x bytes and move buffer pointer x bytesto forward 
//		zero: read more (fill buffer) and call me again (insufficient data in buffer)
//		negative integer (-x): forward x bytes, read more and call me again (I don't use this type of return value)
//		MAX_INT_VALUE: immediately call me again
{
	int filled_len = (*circbuff_ptr3 + BUFF_SIZE - *circbuff_ptr1) % BUFF_SIZE;

	if(we_are_wrong)		//we_are_wrong flag is set if any unhandled case occured. this
		return filled_len;	//is to avoid program crash. note: before setting we_are_wrong flag,
							//the error message must be flushed to stderr

	if(!c2s_handshaked)
	{
		if(filled_len < 68)		//wait for the whole handshake message which is 68 bytes long
			return 0;

		if(!identify(*circbuff_ptr1, filled_len))	//something except a handshake is sent as the first packet -> error
		{
			fprintf(stderr, "WARNING!!! something except handshaking is excahnged as first packet, ");
			fprintf(stderr, "the rest of the traffic of this connection will be ignored.\n");
			we_are_wrong = true;
			return filled_len;
		}

		char temp[50];
		for(int i = 0; i < 20; i++)
			sprintf( temp + 2 * i, "%02x", (unsigned char) *(*circbuff_ptr1 + 48 + i) );
		receiver_id = temp;

		c2s_handshaked = true;
		return 68;
	}

	if(c2s_n_bytes_blindly_fwd > 0)
	{
		if(dbg_info_level & DBG_INF_BFWD)
		{
			fprintf(stderr, "\nBitTorrTrafficIdentifier::c2sFilter: going to forward blindly:\n");
			fprintf(stderr, "\tfilled_len: %d, c2s_n_bytes_blindly_fwd: %d\n\n",
											filled_len, c2s_n_bytes_blindly_fwd);
		}
		if(c2s_n_bytes_blindly_fwd > filled_len)
		{
			c2s_n_bytes_blindly_fwd -= filled_len;
			return filled_len;
		}
		else
		{
			int temp = c2s_n_bytes_blindly_fwd;
			c2s_n_bytes_blindly_fwd = 0;
			return temp;
		}
	}

	if(filled_len < 4)		//we need at least 4 bytes (or probably more) to see what is the message
		return 0;

	unsigned int msg_len = read_int_big_endian(*circbuff_ptr1, circbuff_cptr, *circbuff_ptr3, 0, BUFF_SIZE);

	if(!msg_len)	//keep-alive message
		return 4;	//tell the Conn (caller of this function) to forward the 4-byte message away

	if(filled_len < 4 + 1)	//this "if" is to handle a rare case where only 4 bytes (the msg_len) is passed to this function
		return 0;

	int msg_id = get_byte_from_circbuff(*circbuff_ptr1, circbuff_cptr, *circbuff_ptr3, 4, BUFF_SIZE);

	if(dbg_info_level & DBG_INF_BUFFDATA)
	{
		fprintf(stderr, "\nBitTorrTrafficIdentifier::c2sFilter: \n");
		fprintf(stderr, "\tmsg_id: %d, filled_len: %d, msg_len: %d, c2s_n_bytes_blindly_fwd: %d\n\n",
												msg_id, filled_len, msg_len, c2s_n_bytes_blindly_fwd);
	}

	if(msg_id != 6)	//not a "request" message
	{
		if((unsigned int) filled_len < 4 + msg_len)	//the msg is not received completely, but we want to fwd it, so don't wait
		{
			c2s_n_bytes_blindly_fwd = msg_len - (filled_len - 4);	//this is the length of the remaining part of the message
            if (dbg_info_level & DBG_INF_BFWD)
			{
                fprintf(stderr, "Setting c2s_n_bytes_blindly_fwd to be %d, where filled_len is %d\n\n",
																	c2s_n_bytes_blindly_fwd, filled_len);
            }
			return filled_len;
		}
		else
			return msg_len + 4;
	}

	else			//a "request" message
	{
		if(msg_len != 13)		//invalid BitTorrent request
		{
			cerr << "invalid BitTorrent request, probably the connection is not BitTorrent, ignoring it" << endl;
			we_are_wrong = true;
			return filled_len;
		}

		if((unsigned int) filled_len < 4 + msg_len)	//we need the whole msg to analyze it: 4 is the len of msg_length(=13)
			return 0;

		int piece_index = read_int_big_endian(*circbuff_ptr1, circbuff_cptr, *circbuff_ptr3, 5, BUFF_SIZE);
		int off_in_piece = read_int_big_endian(*circbuff_ptr1, circbuff_cptr, *circbuff_ptr3, 9, BUFF_SIZE);
		int requested_len = read_int_big_endian(*circbuff_ptr1, circbuff_cptr, *circbuff_ptr3, 13, BUFF_SIZE);

		if(dbg_info_level & DBG_INF_REQMSG)
		{
			fprintf(stderr, "\nc2sFilter: request msg: piece_index: %d  off_in_piece: %d  requested_len: %d\n",
																	piece_index, off_in_piece, requested_len);
			//fprintf(stderr, "\t hash: ");
			//for(int i = 0; i < 20; i++) fprintf(stderr, "%02x ", (unsigned char) hash[i]); fprintf(stderr, "\n");
		}

//		here, I use a semi-state-machine to remember not to extract request messages until the piece length is guessed
/*		const int NO_GUESS = 1, TESTING_GUESS = 2, PIECELEN_READY = 3;
		static int state = NO_GUESS;
		static int piecelen_final = -1, piecelen_guess = -1;*/
		switch(state)
		{
			case NO_GUESS:
			{
				if(dbg_info_level & DBG_INF_TI_PIECELEN)
					fprintf(stderr, "\nc2sFilter: NO_GUESS\n");

				int piecelen = Piece_Len_Analyzer->add_request(hash_str, piece_index, off_in_piece, requested_len);
				if(piecelen > 0)
				{
					piecelen_final = piecelen;
					state = PIECELEN_READY;
					if(dbg_info_level & DBG_INF_TI_PIECELEN)
						cerr << "\n\tc2sFilter: piece length " << piecelen << " is guessed for file " << hash_str << endl;
//					go to next case of the switch
				}
				else
				{
#ifdef LOG_REQUESTS
					return 17;
#endif
#ifdef SIMULATION
					return 17;
#endif
/*					char temp[20];
					sprintf(temp, "%08x", piece_index);
					std::string tempstr = temp;
					conn->push_req( new Segment(hash_str + tempstr, off_in_piece, off_in_piece + requested_len) );

//					cache the request
					char *temp_ptr = *circbuff_ptr1 + 13 + 4;
					if(temp_ptr >= circbuff_cptr + BUFF_SIZE)	temp_ptr -= BUFF_SIZE;
					conn->cacheRequest(*circbuff_ptr1, circbuff_cptr, temp_ptr);

					*circbuff_ptr1 += 13 + 4;	//move ptr1 to forward in order to discard (do not forward) the request message
					if(*circbuff_ptr1 >= circbuff_cptr + BUFF_SIZE)
						*circbuff_ptr1 -= BUFF_SIZE;

					return MAX_INT_VALUE;	//call me (this function) again immediately: there might be more messages in the
											//buffer, there is no need to (wait and) read socket again
*/
					return msg_len + 4;	// = 17
				}
			}
			case PIECELEN_READY:
			{
#ifdef SIMULATION
//				close the connection (by sending an illegal message!)
//				write_int_big_endian(*circbuff_ptr1, circbuff_cptr, *circbuff_ptr3, 9, BUFF_SIZE, 500 * 1024 * 1024 + 16384);
//				fprintf(stderr, "\nc2sFilter: sent an illegal piece length (500MB + 16KB), expecting closure for the file ");
//				cerr << hash_str << endl;

				fprintf(stderr, "\nc2sFilter: piece length is guessed, let the traffic pass (just evaluating)\n\n");
				return 17;
#endif
#ifdef LOG_REQUESTS
				return 17;
#endif

				if(dbg_info_level & DBG_INF_TI_PIECELEN)
					fprintf(stderr, "\nc2sFilter: PIECELEN_READY\n");

				if(off_in_piece + requested_len > piecelen_final)
				{
					cerr << "guess for the piece length was wrong (the guess was " <<
							piecelen_final << " while a request addressed offset " <<
							off_in_piece + requested_len << ")! recovering\n\tfile: " << hash_str << endl;
					//TODO: Piece_Len_Analyzer->del_piecelen(hash_str);
					// return 17;	//TODO: WHAT TO DO HERE??????
				    int new_piecelen_final = piecelen_final;
                    while (off_in_piece + requested_len > new_piecelen_final) {
                        new_piecelen_final = new_piecelen_final << 2; 
                    }
					cacheManager->relocate(hash_str, piecelen_final, new_piecelen_final);
                    piecelen_final = new_piecelen_final;
				}

				int off0 = piece_index * piecelen_final + off_in_piece;
				int off1 = off0 + requested_len;
				conn->push_req( new Segment(hash_str, off0, off1, 0, piecelen_final) );

//				remember that this request msg is reported, so that once we saw its response, wo report it as well
				pair < int, pair<int, int> > btreq = make_pair< int, pair<int, int> >
						(  piece_index, make_pair<int, int> (off_in_piece, requested_len)  );
				reported_reqs.insert(btreq);

//				cache the request
				char *temp_ptr = *circbuff_ptr1 + /*request length = */ 13 + 4;
				if(temp_ptr >= circbuff_cptr + BUFF_SIZE)	temp_ptr -= BUFF_SIZE;
				conn->cacheRequest(*circbuff_ptr1, circbuff_cptr, temp_ptr);
				//cout << "request (piece-index: " << piece_index << ", " << "off: " << off_in_piece << ") is cached" << endl;

				*circbuff_ptr1 += 13 + 4;	//move ptr1 to forward in order to discard (do not forward) the request message
				if(*circbuff_ptr1 >= circbuff_cptr + BUFF_SIZE)
					*circbuff_ptr1 -= BUFF_SIZE;

				return MAX_INT_VALUE;	//call me (this function) again immediately: there might be more messages in the
										//buffer, there is no need to (wait and) read socket again
			}
		}//	end switch

		return 17;		//control never reaches here, this is just to avoid compiler's warning
	}//	end else (a "request" message)
}

int BitTorrTrafficIdentifier::s2cFilter(char **circbuff_ptr1, char *circbuff_cptr, char **circbuff_ptr3, char **rawdata_ptr)
//*circbuff_ptr1: address of the first byte of the data in the buffer
//circbuff_cptr:  address of the begining of the buffer memory
//*circbuff_ptr3: address of the end of the data in the buffer (exclusive)
//*rawdata_ptr: must be filled with the pointer to the piece data (if buff contains a "piece" message), and with NULL otherwise
//return value:
//		positivie integer (x): forward x bytes and move buffer pointer x bytesto forward 
//		zero: read more (fill buffer) and call me again (insufficient data in buffer)
//		negative integer (-x): forward x bytes, read more and call me again (I don't use this type of return value)
//		MAX_INT_VALUE: immediately call me again
{
	int filled_len = (*circbuff_ptr3 + BUFF_SIZE - *circbuff_ptr1) % BUFF_SIZE;

	if(we_are_wrong)		//we_are_wrong flag is set if any unhandled case occured. this
		return filled_len;	//is to avoid program crash. note: before setting we_are_wrong flag,
							//the error message must be flushed to stderr

	if(!s2c_handshaked)
	{
		if(filled_len < 68)		//wait for the whole handshake message which is 68 bytes long
			return 0;

		if(!identify(*circbuff_ptr1, filled_len))	//something except a handshake is sent as the first packet -> error
		{
			fprintf(stderr, "WARNING!!! something except handshaking is excahnged as first packet, ");
			fprintf(stderr, "the rest of the traffic of this connection will be ignored.\n");
			we_are_wrong = true;
			return filled_len;
		}

		char temp[50];
		for(int i = 0; i < 20; i++)
			sprintf( temp + 2 * i, "%02x", (unsigned char) *(*circbuff_ptr1 + 48 + i) );
		sender_id = temp;

		s2c_handshaked = true;
		return 68;
	}

	*rawdata_ptr = NULL;					//default is NULL, unless we see a "piece" message (handled below)

//TODO: to Cheng: provide some error failure mechanism (a resp msg corresp to no req msg)
//TODO: what if a wrong (hash incompatible) segment was received as reponse? we will always give it to
//		everyone and they request it again and again -> we proudly think what a popular file we cached!!

	if(s2c_n_bytes_blindly_fwd > 0)
	{
		if(dbg_info_level & DBG_INF_BFWD)
		{
			fprintf(stderr, "\nBitTorrTrafficIdentifier::s2cFilter: going to forward blindly:\n");
			fprintf(stderr, "\tfilled_len: %d, s2c_n_bytes_blindly_fwd: %d\n\n",
											filled_len, s2c_n_bytes_blindly_fwd);
		}
        /* the multi-packet response message is tracked by the Conn object
		if(do_set_rawdata_off)
			*rawdata_ptr = *circbuff_ptr1;
        */

		if(s2c_n_bytes_blindly_fwd > filled_len)
		{
			s2c_n_bytes_blindly_fwd -= filled_len;
			return filled_len;
		}
		else
		{
			int temp = s2c_n_bytes_blindly_fwd;
			s2c_n_bytes_blindly_fwd = 0;
			if(do_set_rawdata_off)
				do_set_rawdata_off = false;

			return temp;
		}
	}

/*	bool two_part = *circbuff_ptr3 < *circbuff_ptr1;	//whether data is splitted over the circular buffer or not
														//(*ptr3==*ptr1 means an empty buffer)
	int len1 = two_part ? (circbuff_cptr + BUFF_SIZE - *circbuff_ptr1) :
							(*circbuff_ptr3 - *circbuff_ptr1);		//length of the first part of data
	int len2 = two_part ? (*circbuff_ptr3 - circbuff_cptr) : 0;		//length of the second part of the data (if exists)
*/
	if(filled_len < 4)		//we need at least 4 bytes (or probably more) to see what is the message
		return 0;

	unsigned int msg_len = read_int_big_endian(*circbuff_ptr1, circbuff_cptr, *circbuff_ptr3, 0, BUFF_SIZE);

	if(!msg_len)	//keep-alive message
		return 4;	//tell the Conn (caller of this function) to forward the 4-byte message away

	if(filled_len < 4 + 1)	//this "if" is to handle a rare case where only 4 bytes (the msg_len) is passed to this function
		return 0;

	int msg_id = get_byte_from_circbuff(*circbuff_ptr1, circbuff_cptr, *circbuff_ptr3, 4, BUFF_SIZE);

	if(dbg_info_level & DBG_INF_BUFFDATA)
	{
		fprintf(stderr, "\nBitTorrTrafficIdentifier::s2cFilter: \n");
		fprintf(stderr, "\tmsg_id: %d, filled_len: %d, msg_len: %d, s2c_n_bytes_blindly_fwd: %d\n\n",
												msg_id, filled_len, msg_len, s2c_n_bytes_blindly_fwd);
	}

	if(msg_id != 7)	//not a "piece" message
	{
		if((unsigned int) filled_len < 4 + msg_len)	//the msg is not received completely, but we want to fwd it, so don't wait
		{
			s2c_n_bytes_blindly_fwd = msg_len - (filled_len - 4);	//this is the length of the remaining part of the message
            if (dbg_info_level & DBG_INF_BFWD)
			{
                fprintf(stderr, "Setting s2c_n_bytes_blindly_fwd to be %d, where filled_len is %d\n\n",
																	s2c_n_bytes_blindly_fwd, filled_len);
            }
			return filled_len;
		}
		else
			return msg_len + 4;
	}

	else			//a "piece" message
	{
		if(msg_len < 9)		//invalid BitTorrent piece message
		{
			cerr << "invalid BitTorrent response, probably the connection is not BitTorrent, ignoring it" << endl;
			we_are_wrong = true;
			return filled_len;
		}

#ifdef LOG_REQUESTS
		Piece_Len_Analyzer->tester_handler.send_keep_alive(hash_str);
		if(dbg_info_level & DBG_INF_RESPMSG)
			fprintf(stderr, "s2cFilter forwarded sth blidnly\n");
		return filled_len;		//no process is needed in s2c: blindly forward everything from sender to receiver
#endif

#ifdef SIMULATION
		return filled_len;		//no process is needed in s2c: blindly forward everything from sender to receiver
#endif

		if(filled_len < 13)		//we need the message parameters (piece_index, off_in_piece, requested_len)
			return 0;

		int piece_index = read_int_big_endian(*circbuff_ptr1, circbuff_cptr, *circbuff_ptr3, 5, BUFF_SIZE);
		int off_in_piece = read_int_big_endian(*circbuff_ptr1, circbuff_cptr, *circbuff_ptr3, 9, BUFF_SIZE);
		int responded_len = msg_len - 9;

		pair < int, pair<int, int> > btreq = make_pair< int, pair<int, int> >
				(  piece_index, make_pair<int, int> (off_in_piece, responded_len)  );
		if(reported_reqs.find(btreq) == reported_reqs.end())	//the req msg associated with the resp msg was not reported
		{
			//cerr << "BitTorrTrafficIdentifier::s2cFilter: the request msg associated with the response msg (" <<
					//piece_index << ", " << off_in_piece << ", " << responded_len <<
					//") was not reported, so the resp msg is not gonna be reported" << endl;
			if((unsigned int) filled_len >= 4 + msg_len)
				return msg_len + 4;
			else
			{
				s2c_n_bytes_blindly_fwd = msg_len - (filled_len - 4);
				if (dbg_info_level & DBG_INF_BFWD)
					fprintf(stderr, "Setting s2c_n_bytes_blindly_fwd to be %d, where filled_len is %d\n\n",
																		s2c_n_bytes_blindly_fwd, filled_len);
				return filled_len;
			}
		}
		reported_reqs.erase(btreq);
		//cerr << "BitTorrTrafficIdentifier::s2cFilter: the response msg (" << piece_index <<
		//		", " << off_in_piece << ", " << responded_len << ") is being reported" << endl;

		if(dbg_info_level & DBG_INF_RESPMSG)
		{
			fprintf(stderr, "BitTorrent piece msg:\n");
			fprintf(stderr, "\t piece_index: %d  off_in_piece: %d\n", piece_index, off_in_piece);
			//fprintf(stderr, "\t hash: ");
			//for(int i = 0; i < 20; i++) fprintf(stderr, "%02x ", (unsigned char) hash[i]); fprintf(stderr, "\n");
	 	}

//		cache the response
		char *temp_ptr = *circbuff_ptr1 + /*response header length = */ 9 + 4;
		if(temp_ptr >= circbuff_cptr + BUFF_SIZE)	temp_ptr -= BUFF_SIZE;
		conn->cacheResponse(*circbuff_ptr1, circbuff_cptr, temp_ptr);

		*circbuff_ptr1 += 13;			//move *ptr1 forward, to discard (do not forward) the msg-len/msg-type/piece-index/off
		if(*circbuff_ptr1 >= circbuff_cptr + BUFF_SIZE)
			*circbuff_ptr1 -= BUFF_SIZE;
		*rawdata_ptr = *circbuff_ptr1;	//let the Conn know where the data starts

		if((unsigned int) filled_len < 4 + msg_len)	//the msg is not received completely, but we want to fwd it, so don't wait
		{
			s2c_n_bytes_blindly_fwd = msg_len - (filled_len - 4);	//this is the length of the remaining part of the message
            if (dbg_info_level & DBG_INF_BFWD)
			{
                fprintf(stderr, "Setting s2c_n_bytes_blindly_fwd to be %d\n\n", s2c_n_bytes_blindly_fwd);
            }
			do_set_rawdata_off = true;
			return filled_len - 13;		//tell the conn to forward the data (but do not forward the "piece" msg header (i.e.
										//msg minus data): to discard this we already have moved circbuff_ptr1 5 bytes forward)
		}
		else
			return msg_len - 9;
	}
}

/*
bool BitTorrTrafficIdentifier::createRequest(char **circbuff_ptr1, char *circbuff_cptr, char **circbuff_ptr3, Segment *seg)
{
	int len = seg->rRange - seg->lRange;
//	.. msg_len (4 bytes) is not changed
//	.. msg code (1 byte) is not changed
//	.. piece index (4 bytes) is not changed
	write_int_big_endian(*circbuff_ptr1, circbuff_cptr, *circbuff_ptr3, 9, BUFF_SIZE, seg->lRange);
	write_int_big_endian(*circbuff_ptr1, circbuff_cptr, *circbuff_ptr3, 13, BUFF_SIZE, len);
}

bool BitTorrTrafficIdentifier::createResponse(char **circbuff_ptr1, char *circbuff_cptr, char **circbuff_ptr3, Segment *seg)
{
	int msg_len = seg->rRange - seg->lRange + 9;	//to be written in the first four bytes of the response
	write_int_big_endian(*circbuff_ptr1, circbuff_cptr, *circbuff_ptr3, 0, BUFF_SIZE, msg_len);
//	.. msg code (1 byte) is not changed
//	.. piece index (4 bytes) is not changed
	write_int_big_endian(*circbuff_ptr1, circbuff_cptr, *circbuff_ptr3, 9, BUFF_SIZE, seg->lRange);
}
*/

bool BitTorrTrafficIdentifier::createRequest(char *buff, int *buff_len, Segment *seg)
{
	int piece_index;
	int len = seg->rRange - seg->lRange;
	int lRange = seg->lRange;

//	either seg->objId is 48 bytes or seg->piecelen is non-zero
	if(seg->objId.length() > 40)	//the piece_index is appended to object ID (piece len was not known)
	{
		strcpy(buff, seg->objId.data() + 40);	//use buff as a temporary buffer!
		sscanf(buff, "%x", &piece_index);
	}
	else	//the piece_index must be inferred
	{
		piece_index = seg->lRange / seg->piecelen;
		lRange = seg->lRange - piece_index * seg->piecelen;
	}

	write_int_big_endian(buff + 0, 13);		//msg_len
	*(buff + 4) = 6;						//msg_id (aka msg code)
	write_int_big_endian(buff + 5, piece_index);
	write_int_big_endian(buff + 9, lRange);
	write_int_big_endian(buff + 13, len);
	*buff_len = 17;

    return true;
}

// build a response message using the most recent request message.
bool BitTorrTrafficIdentifier::createResponse(char *buff, int *buff_len, Segment *seg)
{
	int msg_len = seg->rRange - seg->lRange + 9;	//to be written in the first four bytes of the response
	int piece_index, lRange;

//	either seg->objId is 48 bytes or seg->piecelen is non-zero
	if(seg->objId.length() > 40)	//the piece_index is appended to object ID (piece len was not known)
	{
		strcpy(buff, seg->objId.data() + 40);	//use buff as a temporary buffer!
		sscanf(buff, "%x", &piece_index);
		lRange = seg->lRange;
	}
	else	//the piece_index must be inferred
	{
		piece_index = seg->lRange / seg->piecelen;
		lRange = seg->lRange - piece_index * seg->piecelen;
	}

	write_int_big_endian(buff + 0, msg_len);
    *(buff + 4) = 7; 	   // msg code (1 byte): piece
	write_int_big_endian(buff + 5, piece_index);
	write_int_big_endian(buff + 9, lRange);
    *buff_len = 13;  	 // length 4(len) + 1(id) + 4(idx) + 4(begin)

    return true;
}
