#ifndef _BITTORRTRAFFICIDENTIFIER_H
#define _BITTORRTRAFFICIDENTIFIER_H

#include <string>
#include "TrafficIdentifier.h"
#include "Conn.h"
#include "PieceLenAnalyzer.h"

const int NO_GUESS = 1, PIECELEN_READY = 2;

class BitTorrTrafficIdentifier : public TrafficIdentifier
{
public:
	Conn *conn;
	int BUFF_SIZE;
	char hash[20];//, src_id[20], dest_id[20];	//peer id length and hash length are 20 bytes in BitTorrent
	std::string hash_str;
	bool we_are_wrong;				//this flag is set if any unhandled case occured. this is to avoid program crash,
									//note: before setting we_are_wrong flag, the error message must be flushed to stderr
	bool c2s_handshaked, s2c_handshaked;		//to bypass the handshaking, which is not formatted as c2s/s2c Filter expects
	int c2s_n_bytes_blindly_fwd;	//number of bytes to forward blindly in c2sFilter (to maintain state in the function)
	int s2c_n_bytes_blindly_fwd;	//number of bytes to forward blindly in s2cFilter (to maintain state in the function)
	bool do_set_rawdata_off;		//this is to be used in s2cFilter when the function is called with something
									//the middle of a data chunk (thus without any header)

	int state;		//initially NO_GUESS
	int piecelen_final;

	set< pair< int, pair<int, int> > > reported_reqs;

	PieceLenAnalyzer *Piece_Len_Analyzer;

	BitTorrTrafficIdentifier(char* buff, int buff_len, Conn* conn);
	int c2sFilter(char **, char *, char **);
    int s2cFilter(char **, char *, char **, char **);
    bool createRequest(char *, int *, Segment *);
    bool createResponse(char *, int *, Segment *);

	static bool identify(char *buff, int buff_len)
	{
		return ( buff_len >= 68 && *buff == 0x13 && !strncmp(buff + 1, "BitTorrent protocol", 0x13) );
	}
};

#endif	//#ifndef _BITTORRTRAFFICIDENTIFIER_H
