#ifndef __CACHEMANAGER_H
#define __CACHEMANAGER_H

#include <list>
#include <string>
#include <ext/hash_map>
//#include "pCache.h"
#include "Segment.h"
#include "Item.h"
#include "Cache.h"
#include "LRUCache.h"

// define a hash function for std::string, so that hash_map can be used
namespace __gnu_cxx { template<>
    struct hash< std::string > {
        size_t operator( )( std::string const & s ) const {
            return __stl_hash_string( s.c_str( ) );
        }
    };
}

class CacheManager
{
private:
    __gnu_cxx::hash_map<std::string, Item*> objTable;   // the object table supports lookups
	Cache*	mCache;										// reference to the cache
	pthread_mutex_t cmut;						        // Lock the CacheManager lookup table
    bool merge;
    bool split_copyover;                                // copy over pages during relocation
    long long max_merge;                                // maximum merge size!
	std::set<PageCluster*, DereferenceLess> garbageList;	// page clusters that are going to be recycled
    
public:	
	CacheManager(Cache* _mCache);						                //	constructor
    virtual ~CacheManager();
    std::list<PageCluster*>* lookup(std::string _oid, Segment* _seg);   // search for cached page cluster
                                                                        // caller owns the returned list, and shall
                                                                        // call delete list after it has done with it.
    PageCluster* insert(std::string _oid, Segment* _seg); // return a newly created page cluster
                                                          // the returned object is owned by cache 
                                                          // manager, and will be deleted in remove() 
                                                          // method
                                                          // Note that, if oid is not in the objTable, 
                                                          // a new object will be created as a side-effect
    void remove(std::string _oid, PageCluster* _pc); // Similar, an objet will be removed from the table
                                                     // if it has no page cluster after the delection
    void compact(PageCluster *_pc);                  // page clusters merge?
    void relocate(std::string _oid, long long _old_piece_len, long long _new_piece_len);   // assign a new piece length for BT
	bool contains(std::string);						 // search the hash table for an oid, return True
													 //	if the hash table contain the object
    std::string toString();                          // dump the current state, for debugging purpose
	Item* getItem(std::string _oid);				 // return the requested Item by its _oid
    void cleanup();                                  // check garbage pcList, and free un-used ones
    // The following function is called to get the PMF of page cluster size
    // We round the pc->length to 4 kB to reduce the length of list
    // Nonetheless, this function should be called infrequently due to its high
    // overhead, such as once before exiting
    std::map<long long, long long> get_pc_pmf();
};

#endif   // __CACHEMANAGER_H
