#ifndef __CONN_H
#define __CONN_H

#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>
#include <string>
#include <iostream>
#include <list>

#include "LRUCache.h"
#include "Cache.h"
#include "Util.h"
#include "Segment.h"

#define SOCKET			int

#define INVALID_SOCKET	-1
#define SOCKET_ERROR	-1

class TrafficIdentifier;

class Conn
{
private:
	Cache*  mCache;
    bool byPass;  // By pass flag is raised if the content is
                  // totally un-interested. This can happen when
                  // (1) both admited and chacned lists are empty
                  // (2) both parties are inside our network
                  // (3) both parties are outside our network
    bool swap;    // This flag indicates the role of server and
                  // client should be swapped. This check is done
                  // by matching network mask in the conf file.
    TrafficIdentifier *ti;

	SOCKET  client_socket;
	SOCKET  server_socket;

    int     retriveObject(CacheResponse*, std::string);
    int     cacheDownload(PageCluster*);
    int     serverDownload(Segment*, std::string);
    int     serverDownload(PageCluster*, std::string);

    // a request queue contains a list of unsent requested segments
    std::list<Segment*> reqQueue;
	pthread_mutex_t mut;					// Lock the CacheManager lookup table
public:
    const static int BUFSIZE = 32768;		//	Buffer size

    // inlines to compute read_len, write_len, and update start_pt and end_pt
    // assumes the buffer length is BUFSIZE.
    static inline int recv_length(char* _begin, char* _start, char* _end) {
        if (_end >= _start) {
            if (_start == _begin)
                return (BUFSIZE - (_end - _begin) - 1);
            else 
                return (BUFSIZE - (_end - _begin));
        }
        else return (_start - _end - 1);
    }
    static inline int send_length(char* _begin, char* _start, char* _end) {
        if (_end >= _start) return (_end - _start);
        else return (BUFSIZE - (_start - _begin));
    } 
    // return *end
    static inline char* recv_update_pt(char* _begin, char* _start, char* _end, int ret_val) {
        char *ret = _end + ret_val;
        if (ret - _begin >= BUFSIZE) ret -= BUFSIZE;  
        return ret;
    }
    // return *start
    static inline char* send_update_pt(char* _begin, char* _start, char* _end, int ret_val) {
        char *ret = _start + ret_val;
        if (ret - _begin >= BUFSIZE) ret -= BUFSIZE;
        return ret;
    }
    // get the recv_pointer
    static inline char* recv_start_pt(char* _begin, char* _start, char* _end) {
        return _end;
    }

	Conn(SOCKET, Cache*);
	~Conn();
	void run();  // the main loop

    // request, response buffers, and their cache functions.
    char reqBuf[BUFSIZE];
    char resBuf[BUFSIZE];
    int reqLen;
    int resLen;
    void cacheRequest(char* from, int len);
    void cacheResponse(char* from, int len);
	void cacheRequest(char *circbuff_ptr1, char *circbuff_cptr, char *circbuff_ptr3);
	void cacheResponse(char *circbuff_ptr1, char *circbuff_cptr, char *circbuff_ptr3);
    void push_req(Segment* seg);
    void update_req(long long old_piece_len, long long new_piece_len);
    static void update_all_req(long long old_piece_len, long long new_piece_len);
    void reg();
    void do_tcp_splicing(int fd1, int fd2, char* buf, int len); 
    bool complete_avail(std::list<PageCluster*> *avail_list, Segment *seg);
};

#endif
