#include "GDSCache.h"
#include "CacheManager.h"
#include "pCache.h"
#include <iostream>


//	Calling the parent default constructor
GDSCache::GDSCache() : Cache()
{
}

long long GDSCache::admit(Segment* _seg){
        Item *_item;
		if ((_item = cacheManager->getItem(_seg->objId)) == NULL
			|| (rand() % 1000) < (1000*(_item->freq))/getMaxFreq()) {
            return _seg->rRange - _seg->lRange;
        }
        return 0;
}

void GDSCache::removeItem(PageCluster *_pc)
{
    std::string _oid = _pc->objId;
	if (!q.empty())
		//	find the place of Item
		for (std::set<Item*>::iterator _itemItr = q.begin(); _itemItr != q.end(); ++_itemItr)
			if ((*_itemItr)->objId.compare(_oid) == 0) {
				q.erase(_itemItr);
				break;
			}
}

void GDSCache::abortPC(PageCluster* _pc)
{
	if (!q.empty())
		//	find the place of Item
		for (std::set<Item*>::iterator _itemItr = q.begin(); _itemItr != q.end(); ++_itemItr)
			if ((*_itemItr)->objId.compare(_pc->objId) == 0) {

                //  in addition the PageCluster should be removed from the CacheManager's Hash Table^M
                cacheManager->remove(_pc->objId, _pc);

				if ((*_itemItr)->pcList.empty()) {
					delete *_itemItr;
					q.erase(_itemItr);
				}
				break;
			}
}

//	we don't need to use the servedSize for P2P
void GDSCache::insertItem(PageCluster *_pc, long long servedSize, long long pcServedSize)
{
    Item *_item = _pc->item;
    assert(_item != NULL);

	_item->freq += (10000L * servedSize) / _item->size;
	q.insert(_item);
}

//	Remove a PageCluster from the candidate Item in the queue
void GDSCache::_free()
{
	std::set<Item*>::iterator			_itemItr;
	std::set<Item*>::iterator			tempItr;
	Item*			_item;
	PageCluster*	_pc;
	std::set<PageCluster*, DereferenceLess>::iterator	_pcItr;

	//	For P2P policy select the Item at the bottom of the queue
	_itemItr = q.begin();
	while (size > min_filled_size && _itemItr != q.end()) {
        _item = *_itemItr;
        _pcItr = find_if(_item->pcList.begin(), _item->pcList.end(), victim_pc);
        if (_pcItr != _item->pcList.end()) {  // if we find a good victim
            _pc = *_pcItr;

            //	in addition the PageCluster should be removed from the CacheManager's Hash Table
            cacheManager->remove(_pc->objId, _pc);

            //	if there is no PageCluster remains for the Item
            //	we need to remove the whole Item from the queue
            if (_item->pcList.empty()) {
				tempItr = _itemItr;
                q.erase(tempItr);
                delete(_item);
            }
        }
		else	//	no more eligible page cluster
			_itemItr++;
	}
}
double GDSCache::getMaxFreq()
{
	std::set<Item*>::reverse_iterator	_itemItr;
	_itemItr = q.rbegin();
	if (!q.empty())
		return std::max(1.0, (*_itemItr)->freq);
	return LLONG_MAX;
}

void GDSCache::free_init()
{
    pthread_mutex_lock(&qmut);
	if (q.empty()) {
		std::cout << "WoW: queue is emtpy and yet you want to free the queue!" << std::endl;
		std::cout << "Cache size: " << size << std::endl;
        pthread_mutex_unlock(&qmut);
		return;
	}
	removeItemCandidate	= q.begin();
	removePcCandidate	= (*removeItemCandidate)->pcList.begin();
	initChk = true;
    pthread_mutex_unlock(&qmut);
}

PageCluster* GDSCache::getFreeCandidate()
{
    pthread_mutex_lock(&qmut);
	if (!initChk)
		removePcCandidate++;
	else
		initChk = false;

	while (removeItemCandidate != q.end() && !q.empty()) {
		while (removePcCandidate != (*removeItemCandidate)->pcList.end()) {
			if ((*removePcCandidate)->refCnt <= 0) {	//	we have a candidate to be removed
                pthread_mutex_unlock(&qmut);
				return *removePcCandidate;
			}
			removePcCandidate++;
		}
		removeItemCandidate++;

		//	we assume that we never have an Item with empty PC list
		if (removeItemCandidate != q.end()) removePcCandidate = (*removeItemCandidate)->pcList.begin();
	}
    pthread_mutex_unlock(&qmut);
	return NULL;
}

void GDSCache::freeCurrent()
{
    pthread_mutex_lock(&qmut);
	Item*	_item = *removeItemCandidate;
	std::set<Item*>::iterator tempItr;

	PageCluster *_pc = *removePcCandidate;
	removePcCandidate++;
	initChk = true;


	//	in addition the PageCluster should be removed from the CacheManager's Hash Table
    cacheManager->remove(_pc->objId, _pc);

    //	if there is no PageCluster remains for the Item
    //	we need to remove the Item from the queue
    if (_item->pcList.empty()) {
		tempItr = removeItemCandidate;
		removeItemCandidate++;
		if (removeItemCandidate != q.end()) {
			removePcCandidate = (*removeItemCandidate)->pcList.begin();
		}
        q.erase(tempItr);
        delete(_item);
    }
    pthread_mutex_unlock(&qmut);
}
