#ifndef __GDSCACHE_H
#define __GDSCACHE_H

#include <set>
#include <list>
#include <string>
#include <algorithm>

#include "Cache.h"
#include "Segment.h"
#include "Item.h"

class GDSCache : public Cache
{
public:	
	GDSCache();								//	default constructor
	void free_init();						//	initialize the candidate iterators
	PageCluster* getFreeCandidate();		//	find a candidate PageCluster to be removed
	void freeCurrent();						//	remove the current PageCluster which candidate
											//	iterators point to, from the cache
	long long admit(Segment* _seg);          // called by query(), returns true if we should admit a segment

	void removeItem(PageCluster *_pc);
	void insertItem(PageCluster *_pc, long long servedSize, long long pcServedSize);
	void abortPC(PageCluster* _pc);
	void _free();
    inline bool item_based() {return true; };
    inline int policy() { return GDS; };

private:
	std::set<Item*, DereferenceLess>	q;	//	contains the Item objects
											//	Items are sorted by number of bytes
											//	served for this Item from the cache
	double getMaxFreq();					//	used to get gamma_1 parameter which
											//	is need by p2p admision algorithm
	std::set<Item*>::iterator		removeItemCandidate;//	iterator to the Item candidate
														//	to be removed from the cache in order to
														//	find enough continous disk space.
														//	This should be checked by BlockPool whenever
														//	it wants to find a continious disk space.
	std::set<PageCluster*, DereferenceLess>::iterator	removePcCandidate;	//	same iterator as above except it points
															//	to a PageCluster within an Item
	bool initChk;
};

#endif	//	__GDSCACHE_H
