#include <algorithm>
#include "GnutellaIdentifier.h"
#include "Conn.h"
#include "pCache.h"
#include <sstream>

// constructor makes a connection to the remote side, if such info
// is in payload (non-transparent case) - note: we don't do non-transparent
// proxy cache now - See earlier revision for how to parse CONNECT message
GnutellaIdentifier::GnutellaIdentifier(char* buf, int buf_len, Conn* conn) :
    TrafficIdentifier(buf, buf_len, conn), conn(conn), s2c_remain_data(0)
{
    receiver_id = "receiver"; 
    sender_id = "sender";
}

GnutellaIdentifier::~GnutellaIdentifier() {

}

// The following methods handle two types of messages: GET nd Partial content
// Both messages are terminated with "\r\n\r\n"
int GnutellaIdentifier::c2sFilter(char** _start, char* _begin, char** _end) {
    // find "GET /uri-res/"
    // search for the signature
    int send_len = Conn::send_length(_begin, *_start, *_end);
    int offset = 0;
    memcpy(str_buf + offset, *_start, send_len);  // copy the right hand side portion
    offset += send_len;
    if (*_start > *_end) {  // _start > _end means more data can be copied
        send_len = Conn::send_length(_begin, _begin, *_end);  // 2nd portion
        memcpy(str_buf + offset, _begin, send_len);
        offset += send_len;
    } 
    str_buf[offset] = 0;  // terminate the string
    //std::cout << "c2sFilter, received data: " << str_buf << std::endl;

    char* _head_start = strstr(str_buf, "GET /uri-res/");
    if (_head_start == NULL) return offset;   // no match, pass everything
    if (_head_start != str_buf) return _head_start - str_buf; // not aligned, make it aligned

    char* _head_end = strstr(str_buf, "\r\n\r\n");
    // if we don't see the end return a minus value to ask for receiving more bytes...
    if (_head_end == NULL) {
        return -(_head_start - str_buf);  // read more data
    }
    // copy to connection buffer
    conn->cacheRequest(_head_start, _head_end - _head_start + 4);  // double CRLF

    // parse the request.
    Segment *seg = new Segment();
    strtok(str_buf, ":");
    strtok(NULL, ":");
    seg->objId = (std::string)strtok(NULL, " ");
    char *tmp;
    do {
        tmp = strtok(NULL, "=");
    }
    while (strstr(tmp, "bytes") == NULL); // for bytes = 
    seg->lRange = atoi(strtok(NULL, "-"));
    seg->rRange = atoi(strtok(NULL, "\n")) + 1;     //      using the [a,b) standard
    conn->push_req(seg);

    // filter out the content
    *_start = *_start + (_head_end - _head_start + 4);
    if (*_start >= _begin + Conn::BUFSIZE) 
        *_start -= Conn::BUFSIZE;
 
    return 0;  // ask Conn class to receive more data, then call us again

}

int GnutellaIdentifier::s2cFilter(char** _start, char* _begin, char** _end, char** start_offset) {
    // find "HTTP/1.1 206"
    int send_len = Conn::send_length(_begin, *_start, *_end);
    if (s2c_remain_data > 0) {  // if we have remaining data to send, just send them
        s2c_remain_data -= send_len;
        return send_len;
    }
    // search for the signature, and point the start_offset
    // copy it first, need this for cache command anyway
    int offset = 0;
    memcpy(str_buf + offset, *_start, send_len);  // copy the right hand side portion
    offset += send_len;
    if (*_start > *_end) {  // _start > _end means more data can be copied
        send_len = Conn::send_length(_begin, _begin, *_end);  // 2nd portion
        memcpy(str_buf + offset, _begin, send_len);
        offset += send_len;
    } 
    str_buf[offset] = 0;  // terminate the string
    //std::cout << "s2cFilter, received data: " << str_buf << std::endl;

    char* _head_start = strstr(str_buf, "HTTP/1.1 206");
    if (_head_start == NULL) return offset;   // no match, pass everything
    if (_head_start != str_buf) return _head_start - str_buf; // not aligned, make it aligned

    char* _length_str = strstr(str_buf, "Content-Length: ");
    char* _head_end = strstr(str_buf, "\r\n\r\n");

    // if all three are in buffer, then we are happy, otherwise, return a minus value
    // to ask for receiving more bytes...
    if (_length_str == NULL || _head_end == NULL) {
        return -(_head_start - str_buf);  // read more data
    }
    s2c_remain_data = atoi(_length_str + 16);
    // copy to connection buffer
    conn->cacheResponse(_head_start, _head_end - _head_start + 4);  // double CRLF
    // filter out the content
    *start_offset = *_start + (_head_end - _head_start + 4);
    if (*start_offset >= _begin + Conn::BUFSIZE) 
        *start_offset -= Conn::BUFSIZE;
    *_start = *start_offset;  // skip the message itself
 
    return 0;  // ask Conn class to receive more data, then call us again
}

bool GnutellaIdentifier::createRequest(char *buf, int* buf_len, Segment* seg) {

	std::stringstream local_buf;
	char* cr = buf;

	/*	finding the range fields within a clientRequest,
		modify the fields to the proper value, and send ot to the server socket	*/
    char *tmp;
    tmp = strtok(cr, "=");
    local_buf << tmp;
    while (strstr(tmp, "bytes") == NULL) { // for bytes = 
        tmp = strtok(NULL, "=");
        local_buf << tmp;
    }
	local_buf << "=" << seg->lRange << "-" << seg->rRange - 1 << "\r\n";
	strtok(NULL, "\n");
	local_buf << strtok(NULL, "\0");
	//std::cout << "Sending request: " << local_buf.str() << std::endl;
    *buf_len = local_buf.str().size();
    memcpy(buf, local_buf.str().c_str(), *buf_len);

    return true;
}


// Note: We construct a fresh response and do not use the cached one
bool GnutellaIdentifier::createResponse(char *buf, int* buf_len, Segment* seg) {
	std::stringstream  local_buf;
		
	local_buf	<< "HTTP 200 OK\r\n";
	local_buf	<< "Server: Gnutella\r\n";
	local_buf	<< "Content-type: application/binary\r\n";
	local_buf	<< "Content-length: " << (seg->rRange - seg->lRange - 1) << "\r\n\r\n";

    *buf_len = local_buf.str().size();
    memcpy(buf, local_buf.str().c_str(), *buf_len);
	//std::cout << "Sending response: " << local_buf.str() << std::endl;
    return true;
}


