#include "Item.h"
#include "Util.h"

//	Constructor get the oid and object total size
Item::Item(std::string _oid, long long _size):
   objId(_oid), size(_size), total_hit(0), freq(0)
{
}

//	add a PageCluster into the pcList
void Item::addPC(PageCluster* _pc)
{
    assert(_pc != NULL);
	pcList.insert(_pc);
}

//	Default constructor
Item::~Item()
{
}

//	For debugging purpose
std::string Item::toString()
{
    std::string ret;
    ret.append("ObjId: ").append(objId);
    ret.append(" Size: ").append(Util::stringify(size));
    ret.append(" Total_hit: ").append(Util::stringify(total_hit));
    ret.append(" Freq: ").append(Util::stringify(freq));
    ret.append(" get_num_hit(): ").append(Util::stringify(get_num_hit()));
    // go though the page cluster list
    std::set<PageCluster*, DereferenceLess>::iterator _pos;
    ret.append(" \nPageCluster: ! ");
    for (_pos = pcList.begin(); _pos != pcList.end(); ++_pos) {
	ret.append((*_pos)->toString()).append(" ! ");
    }
    return ret;
}

// return how many times this item has been requested
int Item::get_num_hit() {
    if (pcList.empty() == true) return 0;
    return (*pcList.begin())->num_hit;
}

// we only define less operator for sorted set
// we use left_offset to sort them
bool Item::operator< (const Item &other) const{
    return freq < other.freq;
}

