#ifndef __ITEM_H
#define __ITEM_H

#include <string>
#include <set>

#include "PageCluster.h"
#include "Util.h"

class Item
{
public:
	Item();
	Item(std::string _objId, long long _size);
	void addPC(PageCluster*);		// add a PageCluster to the Item's pcList
									// TODO: if the new PageCluster can be merged with some
									// other PageClusters this function should handle it
									// in a proper way
	~Item();

	std::string objId;				// 32 char
	long long size;					// Cached size in byte
	std::set<PageCluster*, DereferenceLess> pcList;	// Sorted list of allocated page clusters
    // TODO: Ideally, we want to have a multiset here as the priority queue of pc->freq, however
    //       This means we need a separate comparator function here. Do this later.
    std::string toString();         // for debugging purpose

    long long total_hit;    // total number of hit bytes
	double freq;            // this variable can be used in different replacement policies
				            // like LSB, P2P, LFU and all other policies which need to
							// maintain a history
    int get_num_hit();      // returns how many times, this object has been requested
                            // current implementation returns an approximate value 
                            // (how many time the first page cluster was requested) 

    // we define less operator for sorted set
    bool operator< (const Item &other) const;
};

#endif	//	ifndef __ITEM_H
