#ifndef __LFUCACHE_H
#define __LFUCACHE_H

#include <set>
#include <map>
#include <list>
#include <string>
#include <algorithm>

#include "Cache.h"
#include "Segment.h"
#include "Item.h"

class LFUCache : public Cache
{
public:	
	LFUCache();								//	default constructor
	void free_init();						//	initialize the candidate iterators
	PageCluster* getFreeCandidate();		//	find a candidate PageCluster to be removed
	void freeCurrent();						//	remove the current PageCluster which candidate
											//	iterators point to, from the cache

	void removeItem(PageCluster *_pc);
	void insertItem(PageCluster *_pc, long long servedSize, long long pcServedSize);
	void abortPC(PageCluster* _pc);
	void _free();
    inline bool item_based() {return true; };
    inline int policy() { return LFU; };

private:
	std::multiset<Item*, DereferenceLess>	q;	//	contains the Item objects
											//	Items are sorted by number of bytes
											//	served for this Item from the cache
	std::map<std::string, long long> freqHistory;

	std::multiset<Item*>::iterator		removeItemCandidate;//	iterator to the Item candidate
														//	to be removed from the cache in order to
														//	find enough continous disk space.
														//	This should be checked by BlockPool whenever
														//	it wants to find a continious disk space.
	std::set<PageCluster*, DereferenceLess>::iterator	removePcCandidate;	//	same iterator as above except it points
															//	to a PageCluster within an Item
	bool initChk;
	void printQueueSize();
	long long getQueueSize();
};

#endif	//	__LFUCACHE_H
