#include "LRUCache.h"
#include "CacheManager.h"
#include "pCache.h"
#include <iostream>


//	Calling the parent default constructor
LRUCache::LRUCache() : Cache()
{
}

void LRUCache::removeItem(PageCluster *_pc)
{
    q.remove(_pc);
    //stats->lprintf(Stats::INFO, "Cache: detaching item: %s", (*_itr)->objId.c_str()); 
}

void LRUCache::abortPC(PageCluster* _pc)
{
    assert(_pc->refCnt <= 0);
    Item *_item = _pc->item;
    q.remove(_pc);
    //	in addition the PageCluster should be removed from the CacheManager's Hash Table
    cacheManager->remove(_pc->objId, _pc);
    if (_item->pcList.empty()) {
        delete _item;
    }
}

//	insert an item into the queue, for LRU policy
//	we don't need to use the servedSize field
void LRUCache::insertItem(PageCluster *_pc, long long servedSize, long long pcServedSize)
{
	q.push(_pc);
    // stats->lprintf(Stats::INFO, "Cache: pushing back item: %s", _pc->objId.c_str()); 
}

//	Remove a PageCluster from the candidate Item in the queue
void LRUCache::_free()
{
	PageCluster*	_pc;
	PageCluster*	_tmp_pc;
	Item*			_item;

	//	For LRU policy select the page cluster at the bottom of the queue
	for (_pc = q.pop(); size > min_filled_size && _pc != NULL; _pc = _tmp_pc) {
        _item = _pc->item;
        if (victim_pc(_pc) == true) { // can be free.
	        //stats->lprintf(Stats::INFO, "LRUCache: Remove PageCluster with ObjID: %s", _pc->objId.c_str());
	        q.remove(_pc);
            //	in addition the PageCluster should be removed from the CacheManager's Hash Table
            // Important: after the cacheManger->remove, the _pc is freed, and cannot be used!
            // We keep a tmp pointer for iteration purpose
            _tmp_pc = q.get_next(_pc);
            cacheManager->remove(_pc->objId, _pc);
			if (_item->pcList.empty()) {
					delete _item;
            }
        }
    }
}

// free_init() and friend functions should not get the lock, actually, the callers should have
// already got the lock, otherwise we are facing serious race conditions.
void LRUCache::free_init()
{
    pthread_mutex_lock(&qmut);
	if (q.get_size() <= 0) {
		std::cout << "WoW: queue is emtpy and yet you want to free the queue!" << std::endl;
		std::cout << "Cache size: " << size << std::endl;
        pthread_mutex_unlock(&qmut);
		return;
	}
	removePcCandidate = q.pop();
	initChk = true;
    pthread_mutex_unlock(&qmut);
}

PageCluster* LRUCache::getFreeCandidate()
{
    pthread_mutex_lock(&qmut);
	if (!initChk)
		removePcCandidate = q.get_next(removePcCandidate);
	else
		initChk = false;

	while (removePcCandidate != NULL) {
		if (removePcCandidate->refCnt <= 0) {	//	we have a candidate to be removed
            pthread_mutex_unlock(&qmut);
		    return removePcCandidate;
	    }
		removePcCandidate = q.get_next(removePcCandidate);
	}
    pthread_mutex_unlock(&qmut);
	return NULL;
}

void LRUCache::freeCurrent()
{
    pthread_mutex_lock(&qmut);
	PageCluster *_pc = removePcCandidate;
	Item*	_item = _pc->item;
    removePcCandidate = q.get_next(removePcCandidate);
	initChk = true;

	//	in addition the PageCluster should be removed from the CacheManager's Hash Table
    q.remove(_pc);
    cacheManager->remove(_pc->objId, _pc);

    //	if there is no PageCluster remains for the Item
    //	we need to remove the Item from the queue
    if (_item->pcList.empty()) {
        delete _item;
    }
    pthread_mutex_unlock(&qmut);
}

void LRUCache::printQueueSize()
{
	long long _size = 0;

	for (PageCluster *_itr = q.pop(); _itr != NULL; _itr = q.get_next(_itr))
		_size += _itr->length;

	std::cout << "Queue length: " << Util::stringify(q.get_size()) << 
                 " Queue size: " << _size << 
                 "   Cache size: " << size << std::endl;
}
