#include "LSBCache.h"
#include "CacheManager.h"
#include "pCache.h"
#include <iostream>


//	Calling the parent default constructor
LSBCache::LSBCache() : Cache()
{
}

void LSBCache::removeItem(PageCluster *_pc)
{
    std::string _oid = _pc->objId;
	if (!q.empty())
		//	find the place of Item
		for (std::multiset<Item*>::iterator _itemItr = q.begin(); _itemItr != q.end(); ++_itemItr)
			if ((*_itemItr)->objId.compare(_oid) == 0) {
				q.erase(_itemItr);
				break;
			}
}

void LSBCache::abortPC(PageCluster* _pc)
{
	if (!q.empty())
		//	find the place of Item
		for (std::multiset<Item*>::iterator _itemItr = q.begin(); _itemItr != q.end(); ++_itemItr)
			if ((*_itemItr)->objId.compare(_pc->objId) == 0) {

				//	in addition the PageCluster should be removed from the CacheManager's Hash Table
				cacheManager->remove(_pc->objId, _pc);

				if ((*_itemItr)->pcList.empty()) {
					delete *_itemItr;
					q.erase(_itemItr);
				}
				break;
			}
}

void LSBCache::insertItem(PageCluster *_pc, long long servedSize, long long pcServedSize)
{
    Item *_item = _pc->item;
    assert(_item != NULL);

	_item->freq += servedSize/1000L;
	q.insert(_item);
}

//	Remove a PageCluster from the candidate Item in the queue
void LSBCache::_free()
{
	std::multiset<Item*>::iterator			_itemItr;
	std::multiset<Item*>::iterator			tempItr;
	Item*			_item;
	PageCluster*	_pc;
	std::set<PageCluster*, DereferenceLess>::iterator	_pcItr;

	//	For LSB policy select the Item at the bottom of the queue
	_itemItr = q.begin();
	while (size > min_filled_size && _itemItr != q.end()) {
        _item = *_itemItr;
        _pcItr = find_if(_item->pcList.begin(), _item->pcList.end(), victim_pc);
		if (_pcItr != _item->pcList.end()) {  // if we find a good victim
            _pc = *_pcItr;

            //	in addition the PageCluster should be removed from the CacheManager's Hash Table
            cacheManager->remove(_pc->objId, _pc);

            //	if there is no PageCluster remains for the Item
            //	we need to remove the whole Item from the queue
            if (_item->pcList.empty()) {
				tempItr = _itemItr;
				_itemItr++;
                q.erase(tempItr);
                delete(_item);
            }
        }
		else	//	no more eligible page cluster
			_itemItr++;
	}
}

void LSBCache::free_init()
{
    pthread_mutex_lock(&qmut);
	if (q.empty()) {
		std::cout << "error: queue is emtpy and yet you want to free it more!" << std::endl;
		std::cout << "Cache size: " << size << std::endl;
        pthread_mutex_unlock(&qmut);
		return;
	}
	removeItemCandidate	= q.begin();

	removePcCandidate	= (*removeItemCandidate)->pcList.begin();
	initChk = true;
    pthread_mutex_unlock(&qmut);
}

PageCluster* LSBCache::getFreeCandidate()
{
    pthread_mutex_lock(&qmut);
	if (!initChk)
		removePcCandidate++;
	else
		initChk = false;

	while (removeItemCandidate != q.end() && !q.empty()) {
		while (removePcCandidate != (*removeItemCandidate)->pcList.end()) {
			if ((*removePcCandidate)->refCnt <= 0) {	//	we have a candidate to be removed
                pthread_mutex_unlock(&qmut);
				return *removePcCandidate;
			}
			removePcCandidate++;
		}
		removeItemCandidate++;

		//	we assume that we never have an Item with empty PC list
		if (removeItemCandidate != q.end()) removePcCandidate = (*removeItemCandidate)->pcList.begin();
	}
    pthread_mutex_unlock(&qmut);
	return NULL;
}

void LSBCache::freeCurrent()
{
    pthread_mutex_lock(&qmut);
	Item*	_item = *removeItemCandidate;
	std::multiset<Item*>::iterator tempItr;

	PageCluster *_pc = *removePcCandidate;
	removePcCandidate++;
	initChk = true;


	//	in addition the PageCluster should be removed from the CacheManager's Hash Table
    cacheManager->remove(_pc->objId, _pc);

    //	if there is no PageCluster remains for the Item
    //	we need to remove the Item from the queue
    if (_item->pcList.empty()) {
		tempItr = removeItemCandidate;
		removeItemCandidate++;
		if (removeItemCandidate != q.end()) {
			removePcCandidate = (*removeItemCandidate)->pcList.begin();
		}
        q.erase(tempItr);
        delete(_item);
    }
    pthread_mutex_unlock(&qmut);
}
