#ifndef __P2PCACHE_H
#define __P2PCACHE_H

#include <set>
#include <list>
#include <string>
#include <algorithm>

#include "Cache.h"
#include "Segment.h"
#include "Item.h"

class P2PCache : public Cache
{
public:	
	P2PCache();								//	default constructor
	void free_init();						//	initialize the candidate iterators
	PageCluster* getFreeCandidate();		//	find a candidate PageCluster to be removed
	void freeCurrent();						//	remove the current PageCluster which candidate
											//	iterators point to, from the cache
	long long admit(Segment* _seg);          // called by query(), returns how many bytes we should admit 

	void removeItem(PageCluster *_pc);
	void insertItem(PageCluster *_pc, long long servedSize, long long pcServedSize);
	void updateFreq(PageCluster *_pc, long long servedSize, long long pcServedSize);
	void abortPC(PageCluster* _pc);
	void _free();
    inline bool item_based() {return true; };  // the queue is item-based
    inline int policy() { return P2P; };

private:
	std::multiset<Item*, DereferenceLess>	q;	//	contains the Item objects
											    //	Items are sorted by the gamma value!
	std::multiset<Item*>::iterator		removeItemCandidate;//	iterator to the Item candidate
														//	to be removed from the cache in order to
														//	find enough continous disk space.
														//	This should be checked by BlockPool whenever
														//	it wants to find a continious disk space.
	std::set<PageCluster*, DereferenceLess>::iterator	removePcCandidate;	//	same iterator as above except it points
															//	to a PageCluster within an Item
    long long initial_byte;     // admitted amount when the whole object is requested once
    long long expo_factor;      // exponential factor
    long long admit_limit(Item* _item);    // this function returns how many bytes can be admitted
	bool initChk;
    void printQueueSize();
};

#endif	//	__P2PCACHE_H
