#ifndef __PAGEPOOL_H
#define __PAGEPOOL_H

#include <list>
#include <set>
#include "Page.h"
#include "PageCluster.h"

// page pool consists of a few (currently 3) pre-allocated list of buffer
class PagePool {
    private:
	    std::multiset<Page*, DereferenceLess> lBusyPages; // large busy pages
	    std::multiset<Page*, DereferenceLess> mBusyPages; // medium busy pages
	    std::multiset<Page*, DereferenceLess> sBusyPages; // small busy pages
        std::list<Page*> lFreePages; // large free pages
        std::list<Page*> mFreePages; // medium free pages
        std::list<Page*> sFreePages; // small free pages
        float sec_chance_ratio;      // how many pages should go to the second chance buffer?
        std::set<Item*> sec_chance_list; // list of items with something to swap out

    public:
        const static int LPAGE = 1;   // different size of page
        const static int MPAGE = 2;
        const static int SPAGE = 3;

        // size of page pool lists 
        int lPageSize;
        int mPageSize;
        int sPageSize;
        int lPageNum;
        int mPageNum;
        int sPageNum;

        PagePool();   // pre-allocate pages
        ~PagePool();  

        Page* allocate(PageCluster* _pc, int _type); // allocate a page for the caller, this function
                                                     // find the least most recent used page that is
                                                     // not inused (check the ref. cnt. in page cluster)
                                                     // and swap that page out. The pointer to that page 
                                                     // is moved to the tail of its list and is returned.
                                                     // Real callers do not directly use this function, they
                                                     // call the functions in page cluster to achieve that
        void deallocate(Page* _page, int _type);    // moves a page from busy to free list
        void refresh(Page* _page);                  // called when a page is hit again, this page will be
                                                    // moved to the tail of the list.
        std::string toString();                     // dump the info for debugging
        void sec_chance_swapout();                  // swapout all pages on sec_chance_list!
};

#endif // __PAGEPOOL_H
