#ifndef __PIECEANALYZER_H
#define __PIECEANALYZER_H

#include <time.h>
#include <stdio.h>
#include <string>
#include <iostream>
#include <map>
#include <fstream>
#include <pthread.h>
#include <vector>
#include "Util.h"

#ifdef LOG_REQUESTS
#include "TesterHandler.h"
#endif

//#define get_time()	clock()			//in milliseconds
#define TIMEOUT					5 *60*1000	//delete info stored for a file not accessed during last 5 minutes
#define PLEN_ANALYSIS_INTERVAL	500			//in milliseconds
#define MAX(a, b)	((a) > (b) ? (a) : (b))

using namespace std;

//usage of PieceLenAnalyzer class: the c2sFilter function of BitTorrTrafficIdentifier class must call
//PieceLenAnalyzer's add_request for each observed request, and not extract them from messages (just
//say "forward" to the Conn). while adding, if a piee length is guessed it is returned via the add_reqeust
//function, so there is no need to call get_piecelen. the c2sFilter can start extracting request messages
//after having seen the PieceLenAnalyzer's initial guess
class PieceTracker				//holds statistics for a piece being downloaded
{
	map<int, int> *segments;	//pairs of this map are off_in_piece and req_len of requested segments,
								//they are non-overlapping intervals
public:
	int guessed_piecelen;
	long long last_time;		//time of last added request
	int max_delta_time;			//maximum time difference between two consequtive requests
	int max_req_off;			//maximum offset requested ever (within this piece)

	PieceTracker();
	~PieceTracker();
	void add_segment(int off_in_piece, int req_len);

#ifdef SIMULATION
	int n_requests;
#endif
	long long get_n_bytes();	//used by PieceLenAnalyzer::analyze_pieces() for evaluations
	string get_segmap();
//#endif
};

class PieceLenAnalyzer		//bittorrent piece length analyzer
{
	typedef map<int, PieceTracker *> FileStatistics;
	map <string, FileStatistics *> piece_sta;	//hash_id -> map[piece-index -> torre...]
	map <string, long long> last_access;		//hash_id -> last-access-time
	map <string, int> piecelen;					//hash_id -> piece-length
//	for a given file "hash_id", one of the two following conditions hold:
//		1) piecelen[hash_id] is empy and there are some info in last_access[hash_id] and piece_sta[hash_id]
//		2) last_access[hash_id] and piece_sta[hash_id] are empty and there is the piece length in piecelen[hash_id]

	pthread_mutex_t pla_mutex1;		//PieceLenAnalyzer mutex 1: protecting piece_sta and last_access maps
	pthread_mutex_t pla_mutex2;		//PieceLenAnalyzer mutex 2: protecting the "piecelen" map

	PieceLenAnalyzer();
	void delete_file(const string &hash_id);
	int analyze_pieces(const string &hash_id);
	void revisit_garbages();

public:
	~PieceLenAnalyzer();
	void go_down();			//handles desctuction
	int add_request(const string &hash_id, int piece_index, int off_in_piece, int req_len);
//	int get_piecelen(const string &hash_id);	deprecated!
//	void set_piecelen(const string &hash_id, int plen);		deprecated!
	static PieceLenAnalyzer * get_instance()
	{
		#ifndef SIMULATION
		pthread_mutex_lock(&gi_mutex);
		#endif

		static PieceLenAnalyzer *instance = NULL;
		if(instance == NULL)
			instance = new PieceLenAnalyzer();

		#ifndef SIMULATION
		pthread_mutex_unlock(&gi_mutex);
		#endif

		return instance;
	}

#ifdef SIMULATION
	fstream temp_fp;
	int max_size_piece_sta;
#endif
#ifdef LOG_REQUESTS
	pthread_mutex_t thndlr_mutex;
	TesterHandler tester_handler;
#endif

};

#endif		//ifdef __PIECEANALYZER_H
