// This file implements a small status report web server. It also servers an
// auxililary TCP port for our Java applet. Our Java applet reports system
// events in verbose mode.
//
#include <string>
#include <string.h>
#include <iostream>
#include <fstream>
#include <unistd.h>
#include <stdlib.h>
#include <cstdio>
#include <cstdarg>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "Stats.h"
#include "pCache.h"

#define HTTP_PORT "8000"
#define JAVA_PORT "8001"
#define BUFFER_SIZE 1024

void *go_http(void *thread_data)
{
    int connection = (int)thread_data;
    char request_orig[BUFFER_SIZE];
    char request[BUFFER_SIZE];
	char *parse;
	int n;
	unsigned int i;
    std::string response;
	memset(request_orig, '\0', sizeof(request));
	if (read(connection, request_orig, BUFFER_SIZE) > 0)
	{
		request_orig[strlen(request_orig) - 2] = 0;
		parse = strtok(request_orig, " \t\n");
		parse = strtok(NULL, " \t\n");
		snprintf(request, BUFFER_SIZE, "%s%s", parse, 
			(parse[strlen(parse) - 1] == '/') ? "Unknown": "");
        if (0 == strcmp(request, "/applet.html")) {
            response.append("<HTML><HEAD><TITLE>pCache Server Log Applet</TITLE></HEAD>");
            response.append("<BODY><CENTER><P>");
            response.append("<APPLET CODE=\"Log.class\" ARCHIVE=\"Log.jar\" WIDTH=800 HEIGHT=600></APPLET>");
            response.append("</P></CENTER></BODY></HTML>\r\n\r\n");

            for (i = 0; i < response.length(); i += n) {
                n = write(connection, response.c_str(), response.length());
            }

        }
        else if (0 == strcmp(request, "/Log.jar")) {
            struct stat file_stat;
            FILE* fp;
            if ((fp = fopen("../applet/Log.jar", "r")) != NULL) {
                if (lstat("../applet/Log.jar", &file_stat) != 0) {
                    // abort
                    close(connection);
                    fclose(fp);
                    return 0;
                }
                char* result = (char*)malloc(file_stat.st_size);
                fread(result, file_stat.st_size, 1, fp);
                for (i = 0; i < (unsigned int) file_stat.st_size; i += n)
                    n = write(connection, result, file_stat.st_size);
                free(result);
                fclose(fp);
            }
        }
        else {  // by default, we send the server status page
            response.append("<HTML><HEAD><TITLE>pCache Server Status</TITLE></HEAD>");
            response.append("<BODY><CENTER><H1>pCache Proxy Server Status - Version 0.01 </H1><P>");
            response.append("<TABLE BORDER>");
            stats->cons_table(response); // get actual data page
            response.append("<BORDER></P>");
            // refresh button
            response.append("<FORM><INPUT TYPE=\"button\" onClick=\"history.go(0)\" VALUE=\"Refresh\"></FORM>"); 
            response.append("<FORM><INPUT TYPE=\"button\" onClick=\"window.open('applet.html','pCache Proxy Cache','width=800,height=600')\" VALUE=\"Details\"></FORM>");
            response.append("</CENTER></BODY></HTML>\r\n\r\n");

            for (i = 0; i < response.length(); i += n) {
                n = write(connection, response.c_str(), response.length());
            }
        }
    }
    else {
        perror("Http socket write error:");
    }
    close(connection);
	return 0;
}

// We assume the java applet would not push us to hard.
// For example, java applet should never polls us at 1 times/sec rate
// Probably a good rate is around 3 seconds.
void *go_java(void *thread_data)
{
    int connection = (int)thread_data;
    char request[BUFFER_SIZE];
	int n;
	unsigned int i;
    std::string response;
    int accu_read_len = 0;
    int read_len;
    // Note: the mesage length is 9 for now
    int granularity;      // request message
    long long timestamp;  // request message

    do {
        read_len = read(connection, request + read_len, 9 - read_len);
        if (read_len <= 0) { // failed or end of stream
            break;
        }
        accu_read_len += read_len;
        if (accu_read_len >= 9) // read a byte followed by a long integer
	    {
            accu_read_len = 0; // reset the counter
            // parse the message
            granularity = (int)request[0] & 0xff;
            timestamp = ((((long long)request[1]) & 0xff) << 56) |
                        ((((long long)request[2]) & 0xff) << 48) |
                        ((((long long)request[3]) & 0xff) << 40) |
                        ((((long long)request[4]) & 0xff) << 32) |
                        ((((long long)request[5]) & 0xff) << 24) |
                        ((((long long)request[6]) & 0xff) << 16) |
                        ((((long long)request[7]) & 0xff) << 8) |
                        ((((long long)request[8]) & 0xff));
            // printf("Requested level %d, timestamp %lld\n", granularity, timestamp);
            // get a list of the recent logs
            response = stats->getLogs(granularity, timestamp);

            for (i = 0; i < response.length(); i += n) {
                n = write(connection, response.c_str(), response.length());
            }
        }
    } while(true);
    printf("No more data to read, closing socket\n");
    close(connection);
	return 0;
}

// This function starts our httpd that reports general info
void *start_http(void* thread_data) {
    int sd;
    int connect_d;
    socklen_t fromlen;
    struct sockaddr_in sin;
    struct sockaddr_in fsin;

    if ( (sd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
      printf("Unable to initialize http service: cannot get socket\n");
      return NULL;
    }

    // set up reuse address
    int reuse = 1;
    setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(reuse) );

    // bind to all interfaces
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = INADDR_ANY;
    sin.sin_port = htons(atoi(HTTP_PORT));

    // bind to the port.
    if (bind(sd, (const struct sockaddr *) &sin, sizeof(sin)) < 0) {
        printf("Unable to initialize http service: cannot bind\n");
        close(sd);
        return NULL;
    }
 
    // listen for the connection.
    if (listen(sd, 5) < 0) {
        printf("Unable to initialize http service: cannot listen to port\n");
        close(sd);
        return NULL;
    }

    // accept connection.
	while(true){
        pthread_t thread;
        fromlen = sizeof(struct sockaddr);
	    if ( (connect_d = accept(sd, (struct sockaddr *) &fsin, &fromlen)) < 0)
		{
			printf("Unable to initialize http server: cannot accept\n");
			close(sd);
			return NULL;
		}
	else {
		fprintf(stderr, "get a new http connection\n");
	}
        std::string client_addr((char *) inet_ntoa(fsin.sin_addr));
        if (0 == pthread_create(&thread, &pattr, go_http, (void *)connect_d)) {
            pthread_detach(thread);
        }
	}
    close(sd);
    return NULL;
}

void *start_java(void* thread_data) {
    int sd;
    int connect_d;
    socklen_t fromlen;
    struct sockaddr_in sin;
    struct sockaddr_in fsin;

    if ( (sd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
      printf("Unable to initialize java service: cannot get socket\n");
      return NULL;
    }

    // set up reuse address
    int reuse = 1;
    setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(reuse) );

    // bind to all interfaces
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = INADDR_ANY;
    sin.sin_port = htons(atoi(JAVA_PORT));

    // bind to the port.
    if (bind(sd, (const struct sockaddr *) &sin, sizeof(sin)) < 0) {
        printf("Unable to initialize java service: cannot bind\n");
        close(sd);
        return NULL;
    }
 
    // listen for the connection.
    if (listen(sd, 5) < 0) {
        printf("Unable to initialize java service: cannot listen to port\n");
        close(sd);
        return NULL;
    }

    // accept connection.
	while(true){
        pthread_t thread;
        fromlen = sizeof(struct sockaddr);
	    if ( (connect_d = accept(sd, (struct sockaddr *) &fsin, &fromlen)) < 0)
		{
			printf("Unable to initialize java server: cannot accept\n");
			close(sd);
			return NULL;
		}
	else {
		fprintf(stderr, "get a new java connection\n");
	}
        std::string client_addr((char *) inet_ntoa(fsin.sin_addr));
        if (0 == pthread_create(&thread, &pattr, go_java, (void *)connect_d)) {
            pthread_detach(thread);
        }
	}
    close(sd);

    return NULL;
}

// define less operator for sorted set
bool ClientAddr::operator< (const ClientAddr &other) const{
	bool res = client_addr.sin_addr.s_addr < other.client_addr.sin_addr.s_addr;
//	if (client_addr.sin_addr.s_addr == other.client_addr.sin_addr.s_addr)
//		res = client_addr.sin_port < other.client_addr.sin_port;
	return res;
}

// define 'is equal' operator for sorted set
//bool ClientAddr::operator== (const ClientAddr &other) const{
//	return (client_addr.sin_addr.s_addr == other.client_addr.sin_addr.s_addr &&
//		client_addr.sin_port == other.client_addr.sin_port);
//}

ClientAddr::ClientAddr(struct sockaddr_in _addr) : client_addr(_addr)
{
}

ClientAddr::ClientAddr(struct sockaddr _addr) : client_addr(*((struct sockaddr_in*)&_addr))
{
}

ClientStatEntry::ClientStatEntry(std::string _oid) :
	oid(_oid), startTime(Util::get_timestamp()), lastTime(startTime), byte_count(0), total_byte(0)
{
}

LogEntry::LogEntry(long long _timestamp, int _level, std::string _description) :
    timestamp(_timestamp), level(_level), description(_description)
{
}

void Stats::lprintf(int level, const char *lFormat, ...) {
    char lOutput[4096 + 1];
    va_list arglist;
    long long timestamp = Util::get_timestamp();

    // prepare the description
    switch (level) {
        case ERR:
            sprintf(lOutput, "%lld ERR ", timestamp);
            break;
        case WARN:
            sprintf(lOutput, "%lld WARN ", timestamp);
            break;
        case INFO:
            sprintf(lOutput, "%lld INFO ", timestamp);
            break;
        case CLIENT:
            sprintf(lOutput, "%lld CLIENT ", timestamp);
            break;
    }
    int prefix_len = strlen(lOutput);
    va_start(arglist, lFormat);
    vsnprintf(lOutput + prefix_len, 4096 - prefix_len, lFormat, arglist);
    va_end(arglist);

    // remove one \n from the end, if there is any
    if (lOutput[strlen(lOutput)-1] == '\n') {
        lOutput[strlen(lOutput)-1] = 0;
    }

    pthread_mutex_lock(&mut);
	if (level == CLIENT)
		clientLogQueue.insert(LogEntry(timestamp, level, std::string(lOutput)));
    else
		logQueue.insert(LogEntry(timestamp, level, std::string(lOutput)));
    pthread_mutex_unlock(&mut);

    // check whether we need to flush the logs
    flushLogs(false);

    // debug
    //printf("%s\n", lOutput);
}

void Stats::lperror(int level, const char *desc) {
    char lOutput[4096 + 1];
    long long timestamp = Util::get_timestamp();
    
    // prepare the description
    switch (level) {
        case ERR:
            sprintf(lOutput, "%lld ERR %s ", timestamp, desc);
            break;
        case WARN:
            sprintf(lOutput, "%lld WARN %s ", timestamp, desc);
            break;
        case INFO:
            sprintf(lOutput, "%lld INFO %s ", timestamp, desc);
            break;
    }
    int prefix_len = strlen(lOutput);
    char *result = strerror_r(errno, lOutput + prefix_len, 4096 - prefix_len);
    if (result == NULL) {
        std::cout << "Unable to write log messages!" << std::endl;
        return;  // abort
    }
    // copy over the description, because we are using the GNU version (instead
    // of POSIX)
    strncpy(lOutput+prefix_len, result, 4096 - prefix_len);

    // remove one \n from the end, if there is any
    if (lOutput[strlen(lOutput)-1] == '\n') {
        lOutput[strlen(lOutput)-1] = 0;
    }
    pthread_mutex_lock(&mut);
    logQueue.insert(LogEntry(timestamp, level, std::string(lOutput)));
    pthread_mutex_unlock(&mut);
    // check whether we need to flush the logs
    flushLogs(false);
}

void Stats::flushLogs(bool forced) {
	//	pCache stats log
	if (logQueue.size() >= LOG_QUEUE_LEN || forced) {
		std::ofstream file;
		file.open(LOG_FILENAME, std::ios::app | std::ios::out);  
		pthread_mutex_lock(&mut);
		std::set<LogEntry>::iterator pos;
		while((pos = logQueue.begin()) != logQueue.end()) {
			file << pos->description << std::endl;
			logQueue.erase(pos);
		}
		pthread_mutex_unlock(&mut);
		file.close();
	}

	//	Client stats log
	if (clientLogQueue.size() >= LOG_QUEUE_LEN || forced) {
		std::ofstream file;
		file.open(CLIENT_LOG_FILENAME, std::ios::app | std::ios::out);  
		pthread_mutex_lock(&mut);
		std::set<LogEntry>::iterator pos;
		while((pos = clientLogQueue.begin()) != clientLogQueue.end()) {
			file << pos->description << std::endl;
			clientLogQueue.erase(pos);
		}
		pthread_mutex_unlock(&mut);
		file.close();
	}
}

std::string Stats::getLogs(int level, long long timestamp_since) {
    std::string ret;
    pthread_mutex_lock(&mut);
    std::set<LogEntry>::iterator pos;
    // find the first log that has timestamp higher than timestamp_since
    pos = logQueue.upper_bound(LogEntry(timestamp_since, 0, std::string("")));
    for (; pos != logQueue.end(); ++pos) {
        if (pos->level >= level) {  // smaller level means higher precedence
            ret.append(pos->description);
            ret.append(Util::CR_str);
        }
    }
    pthread_mutex_unlock(&mut);
    std::cout << ret;
    return ret;
}

// This function pushes a summary message to the log.txt
void Stats::gen_PR() {
    lprintf(INFO, "PR %lld %lld %lld %lld %lld %f %f %lld %lld %lld %lld", 
            Util::get_timestamp(), request, num_file, num_disk_bytes, 
            num_disk_blocks, disk_byte_utilization, disk_block_utilization, 
            ap_read, ap_write, ap_read_len_in_bytes, ap_write_len_in_bytes);
    lprintf(INFO, "HR %f %lld %lld %lld %lld", 
                  (double)availTraff/(double)totalTraff,
                  totalTraff/1000000LL,
                  availTraff/1000000LL,
                  memTraff/1000000LL,
                  avgLifetime/(1000LL*removedItemCounter+1));
}
// This function pushes some PMF to the log.txt
void Stats::gen_FR(){
    std::map<long long, long long> pmf;
    std::map<long long, long long>::iterator pos;
    pmf = cacheManager->get_pc_pmf();
    for (pos = pmf.begin(); pos != pmf.end(); ++pos) {
        lprintf(INFO, "FR PC_PMF %lld %lld", pos->first, pos->second);
    }
    pmf = blockPool->get_bc_pmf();
    for (pos = pmf.begin(); pos != pmf.end(); ++pos) {
        lprintf(INFO, "FR BC_PMF %lld %lld", pos->first, pos->second);
    }
}

Stats::Stats() :
    request(0),
    high_watermark(0), 
    low_watermark(0),
	totalTraff(0),
	availTraff(0),
	memTraff(0),
	removedItemCounter(0),
	avgLifetime(0),
    small_mem_page_size(0), 
    small_mem_page_number(0), 
    medium_mem_page_size(0),
    medium_mem_page_number(0), 
    large_mem_page_size(0), 
    large_mem_page_number(0), 
    merge(false),
    split_copyover(false),
    disk_block_size(0),
    disk_block_number(0), 
    access_type("unknown"),
    fit_method("unknown"),
    root_dir("unknown"), 
    dev_name("unknown"),
    num_file(0),
    num_disk_bytes(0),
    num_disk_blocks(0),
    disk_byte_utilization(0),
    disk_block_utilization(0),
    ap_read(0),
    ap_write(0),
    ap_read_len_in_bytes(0),
    ap_write_len_in_bytes(0)
{
    // log queue mutex
    pthread_mutex_init(&mut, NULL);
    // start up all services
    pthread_t thread1;
    pthread_t thread2;
    if (0 == pthread_create(&thread1, &pattr, start_http, NULL)) {
        pthread_detach(thread1);
    }
    if (0 == pthread_create(&thread2, &pattr, start_java, NULL)) {
        pthread_detach(thread2);
    }
    // remove the old log file
    remove(LOG_FILENAME);
    remove(CLIENT_LOG_FILENAME);
}

Stats::~Stats() {
    gen_FR();          // create the final report
    flushLogs(true);   // write all leftover messages to the log file
    pthread_mutex_destroy(&mut);
}


// The following function construct the HTML page by filling out all the
// counters
void Stats::cons_table(std::string &ret) 
{
    // general
    ret.append("<TR> <TH COLSPAN=2 BGCOLOR=\"ffff00\"> General Information </TD> </TR>");
    ret.append("<TR><TD>Number of Requests</TD><TD>");
    ret.append(Util::stringify(request));
    ret.append("</TD></TR>");
	ret.append("<TR><TD>Total Number of Threads</TD><TD>");
    ret.append(Util::stringify(total_thread));
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Number of Active Threads</TD><TD>");
    ret.append(Util::stringify(active_thread));
    ret.append("</TD></TR>");

    // cache module
    ret.append("<TR> <TH COLSPAN=2 BGCOLOR=\"ffff00\"> Cache Manager </TD> </TR>");
    //ret.append("<TR><TD>Disk High Watermark</TD><TD>");
    //ret.append(Util::stringify(high_watermark));
    //ret.append("</TD></TR>");
    //ret.append("<TR><TD>Disk Low Watermark</TD><TD>");
    //ret.append(Util::stringify(low_watermark));
    //ret.append("</TD></TR>");
    ret.append("<TR><TD>Byte Hit Rate (\%)</TD><TD>");
    ret.append(Util::stringify(100.0 * (double)availTraff/(double)totalTraff));
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Total Traffic (MB)</TD><TD>");
    ret.append(Util::stringify(totalTraff/1000000));
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Traffic Served from Cache (MB)</TD><TD>");
    ret.append(Util::stringify(availTraff/1000000));
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Traffic Served from Memory (MB)</TD><TD>");
    ret.append(Util::stringify(memTraff/1000000));
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Average Object Lifetime (Sec)</TD><TD>");
    ret.append(Util::stringify(avgLifetime/(1000*removedItemCounter+1)));
    ret.append("</TD></TR>");

	// memory
    ret.append("<TR> <TH COLSPAN=2 BGCOLOR=\"ffff00\"> Buffer Memory Pages </TD> </TR>");
    ret.append("<TR><TD>Small Memory Page Size</TD><TD>");
    ret.append(Util::stringify(small_mem_page_size));
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Small Memory Page Number</TD><TD>");
    ret.append(Util::stringify(small_mem_page_number));
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Medium Memory Page Size</TD><TD>");
    ret.append(Util::stringify(medium_mem_page_size));
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Medium Memory Page Number</TD><TD>");
    ret.append(Util::stringify(medium_mem_page_number));
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Large Memory Page Size</TD><TD>");
    ret.append(Util::stringify(large_mem_page_size));
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Large Memory Page Number</TD><TD>");
    ret.append(Util::stringify(large_mem_page_number));
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Merging Adjacent Page Clusters</TD><TD>");
    if (merge) ret.append("True"); else ret.append("False"); 
    ret.append("</TD></TR>");
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Retain Cached Data During Piece Resizing</TD><TD>");
    if (split_copyover) ret.append("True"); else ret.append("False"); 
    ret.append("</TD></TR>");

    // blocks
    ret.append("<TR> <TH COLSPAN=2 BGCOLOR=\"ffff00\"> Disk Device </TD> </TR>");
    ret.append("<TR><TD>Disk Block Size</TD><TD>");
    ret.append(Util::stringify(disk_block_size));
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Disk Block Number</TD><TD>");
    ret.append(Util::stringify(disk_block_number));
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Disk Access Method</TD><TD>");
    ret.append(access_type);
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Disk Allocation Method</TD><TD>");
    ret.append(fit_method);
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Disk Root Directory</TD><TD>");
    ret.append(root_dir);
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Disk Raw Device Name</TD><TD>");
    ret.append(dev_name);
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Number of Cached Files</TD><TD>");
    ret.append(Util::stringify(num_file));
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Disk Usage in Bytes</TD><TD>");
    ret.append(Util::stringify(num_disk_bytes));
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Disk Usage in Blocks</TD><TD>");
    ret.append(Util::stringify(num_disk_blocks));
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Disk Byte Utilization</TD><TD>");
    ret.append(Util::stringify(disk_byte_utilization));
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Disk Block Utilization</TD><TD>");
    ret.append(Util::stringify(disk_block_utilization));
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Disk Read Operations</TD><TD>");
    ret.append(Util::stringify(ap_read));
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Disk Write Operations</TD><TD>");
    ret.append(Util::stringify(ap_write));
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Disk Read Length in Bytes</TD><TD>");
    ret.append(Util::stringify(ap_read_len_in_bytes));
    ret.append("</TD></TR>");
    ret.append("<TR><TD>Disk Write Length in Bytes</TD><TD>");
    ret.append(Util::stringify(ap_write_len_in_bytes));
    ret.append("</TD></TR>");
}
