#ifndef __UTIL_H
#define __UTIL_H

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <pthread.h>
#include <string>
#include <sstream>
#include <sys/stat.h> // for directory stats
#include <sys/time.h> // for timestamp

//#define EVAL_PieceLenAnalyzer

#define FALSE 0
#define TRUE  1

#define DBG_INF_BFWD			(1 << 1)
#define DBG_INF_BUFFDATA		(1 << 2)
#define DBG_INF_REQMSG			(1 << 3)
#define DBG_INF_TI_PIECELEN		(1 << 4)
#define DBG_INF_RESPMSG			(1 << 5)
#define DBG_INF_MEMFREE			(1 << 6)
#define DBG_INF_PLEN_ANALYSIS0	(1 << 7)
#define DBG_INF_PLEN_ANALYSIS1	(1 << 7)
#define DBG_INF_PLEN_ANALYSIS2	(1 << 8)
#define DBG_INF_PLEN_ANALYSIS3	(1 << 9)
#define DBG_INF_PTRACKER		(1 << 10)
#define DBG_INF_TI_CONSTR		(1 << 11)
#define DBG_INF_PTRACKER_TIME	(1 << 12)
#define DBG_INF_CONN		    (1 << 15)

const unsigned int dbg_info_level = DBG_INF_CONN;
#if 0
DBG_INF_BFWD |
DBG_INF_BUFFDATA |
DBG_INF_REQMSG |
DBG_INF_TI_PIECELEN |
DBG_INF_RESPMSG	|
DBG_INF_MEMFREE	|
DBG_INF_PLEN_ANALYSIS0 |
DBG_INF_PLEN_ANALYSIS1 |
DBG_INF_PLEN_ANALYSIS2 |
DBG_INF_PLEN_ANALYSIS3 |
DBG_INF_PTRACKER |
DBG_INF_TI_CONSTR |	
DBG_INF_PTRACKER_TIME |
DBG_INF_CONN;
#endif

const int MAX_INT_VALUE = ~(1 << (sizeof (int) * 8 - 1));
const unsigned int MAX_LOGGED_PIECES = 50;	//do not keep track of more than 50 pieces for a BitTorrent file
extern pthread_mutex_t gi_mutex;			//to control simultaneous access to PieceLenAnalyzer->get_instance()

// pthread stack size setting
#define THREAD_STACK_SIZE 250000
extern size_t stacksize;
extern pthread_attr_t pattr;


class Util
{
private:
	static long long trace_time;

public:
	static std::string toLower(std::string);
	static std::string tokenizer(std::string*, std::string);

	static void set_tracetime(long long _time)
	{
		trace_time = _time;
	}

	static long long get_tracetime()
	{
		return trace_time;
	}

    template <class T>
    static inline std::string stringify(const T x) {
        std::ostringstream o;
        if (!(o << x)) {  // we failed
            return std::string("unknown format");
        }
        return o.str();
    }
    static std::string CR_str;

    //-----------------------------------------------------------------------------
    // function make_parent_dir
    //
    // purpose  Make sure the parent directory exists for the given path.
    //
    // arguments    1 (const char *) pathname of object a parent is needed for
    //              2 (int) directory access mode
    //
    // returns  (int)  <  0 : error returned by first mkdir()
    //      (int) == -1 : error
    //      (int) ==  0 : parent directory is created
    //      (int) ==  1 : parent directory existed
    //      (int) ==  2 : path does not need a parent
    //
    //-----------------------------------------------------------------------------
    static int mkdirs(const char * arg_path, int arg_mode) {
        struct stat stat_buf;
        char*       slash_ptr;
        char        work_path[ PATH_MAX + 1 ];
        //-- Copy the given path to the work space.
        work_path[ PATH_MAX ] = 0;
        strncpy( work_path, arg_path, PATH_MAX + 1 );

        //-- Must allow at least write and execute permission to owner.
        arg_mode |= ( S_IWUSR | S_IXUSR );

        //-- Step through directory levels and verify access.
        slash_ptr = work_path;
        if ( * slash_ptr == '/' ) { ++ slash_ptr; }  // skip the leading back slash
        while ( * slash_ptr ) {
            if ( * slash_ptr == '/' ) {
                * slash_ptr = 0;
                //-- If the path does not exist, make it.
                if ( stat( work_path, & stat_buf ) < 0 ) {
                    if ( mkdir( work_path, arg_mode ) < 0 ) return -1;
                }
                //-- If it is not a directory, quit now.
                else if ( ! S_ISDIR( stat_buf.st_mode ) ) {
                    return -2;
                }
                * slash_ptr = '/';
            }
            ++ slash_ptr;
        }
        return 0;
    }

    // We define a global elapsed timestamp function, the time stamp is set to
    // zero at bootup time, then calling Util::getTimeStamp() returns a struct
    // timeval
    // Returned relative timestamp is in micro-second
    static void init_misc() {
        gettimeofday(&init_timeval, 0);
        pthread_attr_init(&pattr);
        pthread_attr_getstacksize (&pattr, &stacksize);
        printf("default stack size: %d\n", stacksize); 
        stacksize = THREAD_STACK_SIZE;
        printf("new stack size: %d\n", stacksize); 
        pthread_attr_setstacksize (&pattr, stacksize);
    }
    static struct timeval init_timeval;
    static long long get_timestamp() {
        struct timeval current_timestamp;
        gettimeofday(&current_timestamp, 0);
        return (((long long)current_timestamp.tv_sec) -
                ((long long)init_timeval.tv_sec)) * (long long)1000000 +
                ((long long)current_timestamp.tv_usec) - ((long long)init_timeval.tv_usec);
    }
};

inline char get_byte_from_circbuff(char *ptr1, char *ptr2, char *ptr3, int off, int buff_len)
{
	char *temp = ptr1 + off;
	if(temp >= ptr2 + buff_len)	temp -= buff_len;

	return *temp;
}

inline void put_byte_in_circbuff(char *ptr1, char *ptr2, char *ptr3, int off, int buff_len, char ch)
{
	char *temp = ptr1 + off;
	if(temp >= ptr2 + buff_len)	temp -= buff_len;

	*temp = ch;
}

inline unsigned int read_int_big_endian(char *buff)
{
	return (*buff << 24) | (*(buff + 1) << 16) | (*(buff + 2) << 8) | *(buff + 3);
}

inline void write_int_big_endian(char *buff, int num)
{
	*buff		= (num >> 24) & 0xff;
	*(buff + 1)	= (num >> 16) & 0xff;
	*(buff + 2)	= (num >> 8 ) & 0xff;
	*(buff + 3)	= (num)       & 0xff;
}

inline unsigned int read_int_big_endian(char *ptr1, char *ptr2, char *ptr3, int off, int len)
{
	int a0 = get_byte_from_circbuff(ptr1, ptr2, ptr3, off    , len) & 0xff;
	int a1 = get_byte_from_circbuff(ptr1, ptr2, ptr3, off + 1, len) & 0xff;
	int a2 = get_byte_from_circbuff(ptr1, ptr2, ptr3, off + 2, len) & 0xff;
	int a3 = get_byte_from_circbuff(ptr1, ptr2, ptr3, off + 3, len) & 0xff;

	return (a0 << 24) | (a1 << 16) | (a2 << 8) | a3;
}

inline void write_int_big_endian(char *ptr1, char *ptr2, char *ptr3, int off, int len, int num)
{
	char a0 = (num >> 24) & 0xff;
	char a1 = (num >> 16) & 0xff;
	char a2 = (num >> 8 ) & 0xff;
	char a3 = (num)       & 0xff;

	put_byte_in_circbuff(ptr1, ptr2, ptr3, off    , len, a0);
	put_byte_in_circbuff(ptr1, ptr2, ptr3, off + 1, len, a1);
	put_byte_in_circbuff(ptr1, ptr2, ptr3, off + 2, len, a2);
	put_byte_in_circbuff(ptr1, ptr2, ptr3, off + 3, len, a3);
}

struct DereferenceLess
{
    template <typename PtrType>
    bool operator()(PtrType pT1, PtrType pT2) const {
    return *pT1 < *pT2;
    }
};

struct DereferenceMore
{
    template <typename PtrType>
    bool operator()(PtrType pT1, PtrType pT2) const {
    return *pT1 > *pT2;
    }
};

#endif
