#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <semaphore.h>
#include <sys/types.h> 
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <errno.h>
#include <limits.h>
#include <queue>
#include <vector>

#include "Util.h"
#include "pCache.h"
#include "Conn.h"
#include "libtcpsplice.h"
#include "PieceLenAnalyzer.h"

//	testing the P2P cache or just the connection module
#define _P2P_

//	Trace simulator block size in byte
long long   B_SIZE;
#define		WRITE_BUF_SIZE	1400
char		mData[WRITE_BUF_SIZE];

//	Simul iteration counter
long long	counter	=		0;
#define		counter_print	1000

// Simul request counter, here we count the number of segments
long long   req_counter = 0;
int req_counter_step = 1000;   // default value

//	Event queue parameters
#define		MIN_EVENT	10000
#define		MAX_EVENT	50000

#ifndef _P2P_
#include "pserver.h"
#endif

// ********************************************************
// This section defines global objects that are assume to be
// initialized once. 

// the confs are stored in a map<string, string>
std::map<std::string, std::string> confMap;

// the shared page pool -- use pointer so that we
// can new the object *after* reading the conf.txt
PagePool *pagePool;

// the shared disk block pool
BlockPool *blockPool;

// The cache manager object
CacheManager *cacheManager;

// The stats module
Stats *stats;

//	freq-based policy history parameter
bool keepFreqHistory = false;

long long	availCounter;
long long	segCounter;
long long	reqCounter;
long long	relayCounter;

// whether tcp splicing module is initialized
bool        tcp_splicing;  

// ********************************************************
SOCKET	gListen_Socket;
Cache *mCache;

// This function read conf.txt from current directory, and push configurations
// into confMap.
void init_conf()
{
    std::ifstream conf_in ("conf.txt");
    std::string line;

    if (!conf_in) {
        return;   // don't do anything to the confMap
    }
    if (!conf_in.is_open()) {
        stats->lprintf(Stats::ERR, "Error: Can't open the file: conf.txt\n");
        return;
    }

    while (!conf_in.eof()) {
        getline (conf_in,line);
        if (line.size() > 0 && line.at(0) != '#') { // process it
            std::string::size_type sp_idx = line.find(' ');  // the index of the space char
            if (sp_idx != std::string::npos) {
                std::string key(line.substr(0, sp_idx));
                std::string value(line.substr(sp_idx + 1));
                //  std::cout << "Inserting key: " << key << " value: " << value << std::endl;
                confMap.insert(std::make_pair(key, value));
            }
        }
    }
    conf_in.close();
}

void simul_dl(Segment* reqSeg)
{
	long long current = reqSeg->lRange; // the global offset, -1 indicates un-initialized
    bool admitted = false;				// should we copy the content into memory
    PageCluster* pc = NULL;				// current page cluster
#if 0 // disable memcpy to test disk throughput
	char read_buf[20000];				// read buffer from page clusters
	long long read_length = 17000;			// how much we should read for test
#endif

	//	requesting the segment from the cache
	CacheResponse* mResponse = mCache->query(reqSeg);

	std::list<PageCluster*>::iterator availableIter;
	std::list<PageCluster*>::iterator requestedIter;

	//	update the counter log
	++segCounter;
	availCounter+= mResponse->availableSegment->size();
	reqCounter	+= mResponse->requestedSegment->size();
	//	relayCounter will be updated later on

    // std::cout << "Bear " << mCache->getSize()/1000000 << " MB" << " " << reqSeg->toString() << std::endl;

	availableIter = mResponse->availableSegment->begin();
	requestedIter = mResponse->requestedSegment->begin();

//std::cout << "Getting into the request handler with current: " << current << " and reqseg: " << reqSeg->toString() << std::endl;
//std::cout << "Cache response: " << mResponse->toString() << std::endl;

//int iii = 0;
    // find a new subsegment
	while (current < reqSeg->rRange) {
        // the left range is always the current, but the right range is the
        // closest boundary of (*availableIter)->getLeft(), ->getRight(),
        // (*requestedIter)->getLeft(), ->getRight().  In the cases, where
        // we pick the left boundary, no in-memory buffer read/write is
        // needed.
        // If we pick (*availableIter)->getRight(); then we should pump the
        // data immediate, and find next subReSeg
        // Finally, if we pick (*requestedIter)->getRight(), we should make
        // a request, and copy data into memory
        bool avail;   // whether the subsegment is in our cache
        long long next_end = LLONG_MAX;
        admitted = false;
        avail = false;
        if (availableIter != mResponse->availableSegment->end() &&
            (*availableIter)->getRight() <= next_end) {
            next_end = (*availableIter)->getRight();
            admitted = false;
            avail = true;           // play out from our cache
            pc = (*availableIter);
        }
        if (requestedIter != mResponse->requestedSegment->end() &&
            (*requestedIter)->getRight() <= next_end) {
            next_end = (*requestedIter)->getRight();
            admitted = true;
            avail = false;          // copy into our cache
            pc = (*requestedIter);
        }
        if (availableIter != mResponse->availableSegment->end() &&
            (*availableIter)->getLeft() > current && 
            (*availableIter)->getLeft() < next_end) {
            next_end = (*availableIter)->getLeft();
            admitted = false;
            avail = false;          // relay
            pc = NULL;
        }
        if (requestedIter != mResponse->requestedSegment->end() &&
            (*requestedIter)->getLeft() > current && 
            (*requestedIter)->getLeft() < next_end) {
            next_end = (*requestedIter)->getLeft();
            admitted = false;
            avail = false;          // relay
            pc = NULL;
        }
        next_end = std::min(next_end, reqSeg->rRange);

        if (++req_counter % req_counter_step == 0) {
//			std::cout << "\tRequest Counter: " << req_counter << ": \t";

			// generate a PR report
            stats->gen_PR();
        }
        // update the req_counter to stats module
        stats->request = req_counter;

		// if we have the current part, send it to the client
		if (admitted == false && avail == true) {
#if 0 // disable memcpy to test disk throughput
			pc->read(0, MIN (pc->length, read_length), read_buf);
			read_buf[MIN (pc->length, read_length)] = 0;
			//printf("%s\n", read_buf);
#endif
			pc->decRefCnt();
			availableIter++;

            // send from current to next_end
            // no need to create subSeqSeg here, we handle the request right away
//			std::cout << "Serve from  cache: " << reqSeg->objId << " " << current << ":" << next_end << std::endl;
//            if (next_end >= reqSeg->rRange) { // means we are are not in with the current segment
//                delete reqSeg;
//                reqSeg = NULL;
//            }
			current = next_end;
		}
		if (admitted == true && avail == false) {
			//	write data into the cache buffer
#if 0 // disable memcpy to test disk throughput
			long long i;
			for (i = current; i < next_end - WRITE_BUF_SIZE; i += WRITE_BUF_SIZE) {
				pc->write(i - pc->offset, WRITE_BUF_SIZE, mData);
			}
			if (i < next_end)
				pc->write(i - pc->offset, next_end - i, mData);
#endif

			pc->received();		//	The PageCluster is ready to store on the disk
			pc->decRefCnt();
			
			requestedIter++;
//			std::cout << "Storing in  cache: " << reqSeg->objId << " " << current << ":" << next_end << std::endl;
			current = next_end;
		}
		if (admitted == false && avail == false) {
//			std::cout << "Relay the segment: " << reqSeg->objId << " " << current << ":" << next_end << std::endl;
			++relayCounter;
			current = next_end;
        }
//		if (reqSeg != NULL && current >= reqSeg->rRange) { // means we are done with the current segment
//            delete reqSeg;
//            reqSeg = NULL;
//        }
	}
#if 0
// Cheng, 08/01/05, we remove this check for COSS. COSS system may evict a segment while not in query(...) function
//                  which is different from the original design, and will fail here!
	assert(availableIter == mResponse->availableSegment->end());
	assert(requestedIter == mResponse->requestedSegment->end());
	for (availableIter = mResponse->availableSegment->begin();
			availableIter != mResponse->availableSegment->end(); availableIter++)
    {
        assert((*availableIter)->refCnt == 0);
    }
	for (requestedIter = mResponse->requestedSegment->begin();
			requestedIter != mResponse->requestedSegment->end(); requestedIter++)
    {
        assert((*requestedIter)->refCnt == 0);
    }
#endif

//std::cout << "Getting out of the request handler with current: " << current << std::endl;
	if (reqSeg) {
		delete reqSeg;
//        reqSeg = NULL;
    }
	if (mResponse) {
		delete mResponse;
//		mResponse = NULL;
	}
}

void simul_event(char* in_file)
{
	std::string		line;
	std::ifstream	in;
	in.open(in_file);

	int				i, j;
	long long		r1, l1, pos, lRange, rRange;
	std::string		time;
	std::string		oid;
	long long		size;
	std::string		ip;
	//	number of requests should be read from the trace file
	long long	itrCounter = atoll(confMap[std::string("TRACE_REQUEST_NUM")].c_str());
	//	check if the aborted download is enable
	bool			abort_download = (atol(confMap[std::string("ABORT_DOWNLOAD")].c_str()) == 1)?true:false;
	bool			abort = false;

	//	initialize the random seed
	srand(0);

	std::priority_queue<Segment*, std::vector<Segment*>, DereferenceMore> event;
	long long delay;
	long long trace_time, completion_time;

	if (!in.is_open()) {
		std::cout << "Error: opening simul file." << std::endl;
		return;
	}

	//	filling the sample data buffer
//	for (i = 0; i < WRITE_BUF_SIZE; i++)
//		mData[i] = (char)(i%10) + 48;

	while (! in.eof()) {
		//	maximum number of iterations
		if (++counter > itrCounter) {
	        // std::cout << "Finished " << counter << " iteration." << std::endl;
            break;
        }

		//	Cache warm up
//		if (counter == 10000) {
//			stats->totalTraff = 0;
//			stats->availTraff = 0;
//			stats->memTraff = 0;
//		}
//		++counter;
//		if (counter % counter_print == 0) {
//			std::cout << "Iteration #" << counter << ": \t";
//			std::cout << mCache->getSize()/1000000 << " MB" << std::endl;
//		}
		getline(in, line);
        try {   // catch the out-of-range substr exception
            i = 0;
            while (line[i] != '\t') {
                i++;
            }
            if (i != 32)
                continue;
            oid = line.substr(0, i);
            j = ++i;
    
            while (line[i] != '\t') {
                i++;
            }
            ip = line.substr(33, i - 33);
            j = ++i;
    
            while (line[i] != '\t') {
                i++;
            }
//            size = 1 + atoll(line.substr(j, i - j).c_str())/25LL;
            size = atoll(line.substr(j, i - j).c_str());
            if (size < 1)
                continue;

			j = ++i;
    
            while (line[i] != '\t' && line[i] != '\0') {
                i++;
            }
            time = line.substr(j, i - j);
        }
        catch(...) { // catch all exceptions 
            std::cout << "Unable to parse log-line: " << line << " skip it!" << std::endl;
            continue;
        }
		//	with 66% prbability a download session might be aborted
		if (abort_download) {
			if ((rand() % 100) < 66)
				abort = true;
			else
				abort = false;
		}

		//	The line bandwidth can be between 5 KB/s to 0.5 MB/s
		delay = 5 + (rand() % 496);
		completion_time = size/(3*delay);
        trace_time = atoll(time.c_str());

		//	set the current trace time
		Util::set_tracetime(trace_time);

		//	---- Segment ----
		l1 = 0;
		r1 = size - 1;

		pos = l1;
		while (pos < r1) {
			lRange = pos;

			if (pos + B_SIZE - 1 > r1)
				pos = r1;
			else
				pos	+= B_SIZE - 1;

			rRange = ++pos;

			//	if this request is suposed to be aborted,
			//	each requested segment has 1/2 chance to be aborted
			if (!abort || (rand() % 100) < 50)
				event.push(new Segment(oid, lRange, rRange, trace_time + (rand() % 100)*completion_time));
		}

		//	if the event queue is full
		if (event.size() > MAX_EVENT) {
//			std::cout << "reading the event queue..." << std::endl;
			while (event.size() > MIN_EVENT) {
				simul_dl(event.top());
				event.pop();
			}
		}
	}

	//	reading the remaining requests in queue
	while (!event.empty()) {
		simul_dl(event.top());
		event.pop();
	}

    in.close();
}

int startServer(int port)
{
	sockaddr_in local;
	SOCKET listen_socket;

	local.sin_family = AF_INET;
	local.sin_addr.s_addr = INADDR_ANY;
	local.sin_port = htons(port);

	listen_socket = socket(AF_INET, SOCK_STREAM, 0);
	if(listen_socket == INVALID_SOCKET) {
		stats->lprintf(Stats::ERR, "Error in New a Socket.");
		return -2;
	}

    /* tproxy */
    int hack_value = 1;
    int rr = setsockopt(listen_socket, SOL_IP, IP_TRANSPARENT, &hack_value, sizeof(hack_value)); 
    if (rr == -1) {
        stats->lperror(Stats::ERR, "");
	    return -2;
    }

    // enable address reuse, use the same variable used in tproxy
    rr = setsockopt(listen_socket, SOL_SOCKET, SO_REUSEADDR, &hack_value, sizeof(hack_value) );
    if (rr == -1) {
        stats->lperror(Stats::ERR, "");
	    return -2;
    }

	if(::bind(listen_socket,(sockaddr *)&local,sizeof(local))!=0) {
		stats->lprintf(Stats::ERR, "Error in Binding socket."); 
		return -3;
	}

	if(::listen(listen_socket, SOMAXCONN)!=0) {
		stats->lprintf(Stats::ERR, "Error in Listen.");
		return -4;
	}
	gListen_Socket = listen_socket; 

    // initialize TCP splice engine if the option is on
    if (0 == confMap[std::string("TCP_SPLICING")].compare("0")) {
        // no splicing 
        printf("TCP splicing engine is disabled.\n");
        tcp_splicing = false;  
    }
    else {
        if (tcp_splice_start() < 0) { // if startup failed
            stats->lprintf(Stats::ERR, "Unable to initialize TCP splicing engine, all bypass connections are handle in applicationlayer!\n");
            printf("Unable to initialize TCP splicing engine, all bypass connections are handle in applicationlayer!\n");
            tcp_splicing = false;  
        }
        else {
            stats->lprintf(Stats::INFO, "TCP splicing engine is initialized!\n");
            printf("TCP splicing engine is initialized!\n");
            tcp_splicing = true;  
        }
    }

	//	start a new tread
	pthread_t thread;
	if(pthread_create(&thread, &pattr, UserToProxyThread, NULL) != 0)
		return -5;
    else
        pthread_detach(thread);

	return 1;
}

int closeServer()
{
	close(gListen_Socket);
	return 1;
}

/*	Setup chanel and read data from local , send data to remote	*/
void *UserToProxyThread(void* pParam)
{
	SOCKET		client_socket;
	sockaddr_in	from;
	int			fromlen;
	pthread_t	mThread;

	fromlen = sizeof(from);
	client_socket = accept(gListen_Socket, (struct sockaddr*)&from, (socklen_t*)&fromlen);

	//	increament the number of threads
	stats->total_thread++;
	stats->active_thread++;

	//	fork a thread
	if(pthread_create(&mThread, &pattr, UserToProxyThread, NULL) != 0) {
		return NULL;
    }
    else {
        pthread_detach(mThread);
    }

	if(client_socket == INVALID_SOCKET) {
        stats->lperror(Stats::ERR, "Error in accept");
		return NULL;
	}

	Conn conn (client_socket, mCache);
    conn.reg(); // register conn object to the global connection table
	conn.run();

	//	decrement number of active threads
	stats->active_thread--;

	return 0;
}

pthread_mutex_t gi_mutex;			//to control simultaneous access to delete_file()

void flush_logs(){
    delete stats;
}

int main(int argc, char *argv[])
{
	int port;
    // setup the global timstamp and other setting (like pthread attr)
    // Use this to print timestamp: 
    //      std::cout << Util::get_timestamp() << std::endl;
    Util::init_misc();

    // init the stats module
    stats = new Stats();
    // install an at exit function to flush leftover logs
    atexit(flush_logs);

    // init configurations..
    init_conf();

	//	check the input arguments
/*	TODO
	for (int i = 1; i < argc; i++) {
		if (argv[i][0] != '-') {
			print_help();
			exit(1);
		}
		switch (argv[i][1]) {
			case 'r': rp = 
			default	: print_help();
				break;
		}
	}
*/
	//	cache capacity
	if (argc > 3)
		confMap[std::string("BLOCK_NUM")] = std::string(argv[3]);

	// The print steps for request counter
    std::stringstream req_counter_step_str(confMap[std::string("REQUEST_COUNTER_STEP")]);
    req_counter_step_str >> req_counter_step;
    // The generated request size
    std::stringstream req_size_str(confMap[std::string("REQUEST_SIZE")]);
    req_size_str >> B_SIZE;

    // init block pool subsystem
    blockPool = new BlockPool();

    // init page pool subsystem
    // Note: the init order is important because page pool uses some block pool parameters
    pagePool = new PagePool();

	//	initializing the log variables
	segCounter = 0;	availCounter = 0;
	reqCounter = 0;	relayCounter = 0;	

	if (argc >= 1) {
		/* ... = new CacheManager(Replacement Policy, size);
			where RP can either LRU_CACHE, LFU_CACHE or etc.,
			and size is size of the cache in number of bytes	*/
        	if ((argc > 1 && blockPool->accesorType == 6) || // force to use COSS 
		    (argc > 1 && strcmp(argv[1], "coss") == 0)) 
			mCache = new COSSCache();
		else if (argc > 1 && strcmp(argv[1], "lru") == 0)
			mCache = new LRUCache();
		else if (argc > 1 && strcmp(argv[1], "lfu") == 0)
			mCache = new LFUCache();
		else if (argc > 1 && strcmp(argv[1], "lsb") == 0)
			mCache = new LSBCache();
		else if (argc > 1 && strcmp(argv[1], "p2p") == 0)
			mCache = new P2PCache();
		else {
			std::cout << "Warning: the first argument should be the policy type. Using LRU for now." << std::endl;
			mCache = new LRUCache();
		}

		// init cache manager subsystem
		cacheManager = new CacheManager(mCache);
		
		if (argc > 2) {
			//	running in connection mode
			if (isdigit(argv[2][0]))
				port = atoi(argv[2]);
			else {
			//	running in simul mode
				std::cout << "Run in Trace Simulation mode. No connection handling." << std::endl;

				simul_event(argv[2]);

				std::cout << "Current time: " << Util::get_timestamp() << std::endl;
				std::cout << "Ava: " << 1000*availCounter/segCounter << ", Req: " << 1000*reqCounter/segCounter << ", Rel: " << 1000*relayCounter/segCounter << std::endl;

				exit(0);
			}
		} else
			//	default port number
			port = 7072;
	}
	else {
		port = 7072;
    }

	pthread_mutex_init(&gi_mutex, NULL);		//for get_instance() method of PieceLenAnalyzer class

    // init the server daemon
	startServer(port);

	std::cout << "P2P proxy server is listening on port: " << port << ". To exit press 'q' and Enter." << std::endl;

	PieceLenAnalyzer *PLenAnalyzer = PieceLenAnalyzer::get_instance();	//this get_instance is only to have
//								some reference to the unique instance of PieceLenAnalyzer to delete it at last

	while(TRUE)	
		if(getchar()=='q') break;

    // shut down the server daemon
	closeServer();
    // stop the tcp splicing engine
#if 0 // ignore this for now, the tcp_splice module still crashes for unknown reasons
    if (tcp_splicing == true) {
        tcp_splice_stop();
    }
#endif
	PLenAnalyzer->go_down();	//not deleting PieceLenanalyzer is because some threads may still be working with it
//	delete PLenAnalyzer;
	pthread_mutex_destroy(&gi_mutex);

	return 1;
}
